//
//***************************************************************************
// Name    : USBHID.bas                                                     *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 10/01/2007                                                     *
// Version : 1.4 updated for new stack version                              *
//         : 1.1 ISR now uses USB interrupts                                *
//         : Renamed 'Connected' to 'Attached'                              *
//         : 1.0 Release                                                    *
// Notes   : This HID module provides a simple way for you to transfer      *
//         : blocks of data to and from your PC using the USB HID class.    *
//***************************************************************************
//
module HID

// import USB modules...
#option USB_USE_HID = true
include USB_DESCRIPTOR
include "usbsfr.bas"
include "usbmem.bas"
include "usbdefs.bas"
include "usbsystem.bas"
include "system.bas"        // for clrWDT

// HID memory requirements
// HID_USED_RAM = (EP_SIZE * (MAX_EP_NUMBER+1) * 2) +       ' = 4 * (3+1) * 2 = 32
//                sizeof(structure CTRL_TRF_SETUP) +        ' = 8
//                sizeof(structure CTRL_TRF_DATA) +         ' = 8
//                HID_REPORT_IN_SIZE +                      ' = 64 (HID_REPORT_BYTES_IN)
//                HID_REPORT_OUT_SIZE                       ' = 64 (HID_REPORT_BYTES_OUT)
//
// so HID_USED_RAM = 176 bytes ($B0)
// If there's a minimum of 256 bytes of USB ram, that leaves 256-176=80 bytes available
// for extended ram even on the smallest of parts
//
#variable _HID_USED_RAM = 176

// this option controls if Initialize() is called automatically
#option USB_AUTO_INITIALIZE = true

// ISR priority
const
    PRIORITY_LEVEL = USB_SERVICE_PRIORITY

//
// EXTENDED_RAM support (see usbmem.bas)
// Public buffers - these are just mapped onto USB ExtendedRAM space, thus saving
// program RAM. You can use a large 256 byte buffer, which can be useful
// for transferring large blocks of data using ReadArray() and WriteArray().
// The RX and TX buffers match the HID report size, and are typically use with
// ReadReport() and WriteReport(). However, they can also be used with ReadArray()
// and WriteArray(). Note that these buffers are only available for USB devices
// that have extended ram defined
//
#if (USB_EXTENDED_RAM_SIZE > 0)
#if ((HID_REPORT_IN_SIZE + HID_REPORT_OUT_SIZE) > USB_EXTENDED_RAM_SIZE)
  #error "USB_EXTENDED_RAM_SIZE too small"
#endif
public const
    EXT_RAM_ADDRESS = USB_EXTENDED_RAM_BASE,
    TX_REPORT_RAM = EXT_RAM_ADDRESS,
    RX_REPORT_RAM = EXT_RAM_ADDRESS + HID_REPORT_IN_SIZE

public dim
    Buffer(USB_EXTENDED_RAM_SIZE)    as byte absolute EXT_RAM_ADDRESS,
    TXReport(HID_REPORT_IN_SIZE)     as byte absolute TX_REPORT_RAM,
    RXReport(HID_REPORT_OUT_SIZE)    as byte absolute RX_REPORT_RAM
#endif      // USB_EXTENDED_RAM_SIZE

public dim
    Attached as USBSystem.Attached

//
//**************************************************************************
//Name    : Service                                                        *
//Purpose : Service a USB connection (private)                             *
//**************************************************************************
//
inline sub _service()
    USBCheckBusStatus()
    USBDriverService()
    ClrWDT()
end sub

//
//**************************************************************************
//Name    : USBService (was private NoISRService)                          *
//Name    : Service                                                        *
//Purpose : Service a USB connection. You need to call this routine        *
//        : periodically from your main program to maintain a connection   *
//        : (every 1 ms or so). You DO NOT need to call this routine if    *
//        : you have enabled the USB interrupt handler, since the routine  *
//        : will be optimized away                                         *
//**************************************************************************
//
public inline sub USBService()
  #if (USB_SERVICE = false)     // only call _service if polled mode
    _service()
  #endif
end sub

//
//**************************************************************************
//Name    : USB_intr_handler (PRIVATE)                                     *
//Purpose : USB isr handler                                                *
//**************************************************************************
// this routine takes ~20us at 48MHz, but since in most cases interrupts are 
// generated back to back the effective time is ~40us. 
// they occur at 1ms intervals, but that is host dependant
//
#if (USB_SERVICE)
interrupt USB_intr_handler(PRIORITY_LEVEL)
    // check for USB intr request
    if ((PIR2Bits.USBIF = 1) and (PIE2Bits.USBIE = 1)) then
        PIR2Bits.USBIF = 0
        save(FSR0, FSR1, _service)
        _service()
        restore
    endif
end interrupt
#endif

//
//**************************************************************************
//Name    : EnableISR                                                      *
//Purpose : Enable the ISR to service USB connection                       *
//**************************************************************************
//
public inline sub EnableISR()
  #if (USB_SERVICE)
    PIE2Bits.USBIE = 1
  #endif
end sub

//
//**************************************************************************
//Name    : DisableISR                                                     *
//Purpose : Disable the ISR that services the USB connection               *
//**************************************************************************
//
public inline sub DisableISR()
  #if (USB_SERVICE)
    PIE2Bits.USBIE = 0
  #endif
end sub

//
//**************************************************************************
//Name    : WaitForTX (was PRIVATE)                                        *
//Purpose : Blocks processing until USB TX buffer is available             *
//**************************************************************************
//
public sub WaitForTX()
    while ((mHIDTxIsBusy <> 0) and Attached)
        USBService()
    end while
end sub
{
// note: do not call this routine if USB_SERVICE=true and interrupts are enabled
sub WaitForTX()
    while ((mHIDTxIsBusy <> 0) and Attached)
        _service()
    end while
end sub
}

//
//**************************************************************************
//Name    : DataAvailable                                                  *
//Purpose : Returns true if RX data is available, false otherwise          *
//**************************************************************************
//
public function DataAvailable() as boolean
    DisableISR()
    result = (mHIDRxIsBusy = 0)
    EnableISR()
end function

//
//**************************************************************************
//Name    : ReadReport                                                     *
//Purpose : Read a HID report into RXReport buffer                         *
//**************************************************************************
//**************************************************************************
//Name    : WriteReport                                                    *
//Purpose : Write the TXReport buffer                                      *
//**************************************************************************
// note: these routines require ExtendedRam
// see main_hid.bas for an example of using ReadReport()/WriteReport()
// without ExtendedRam
//
#if (USB_EXTENDED_RAM_SIZE > 0)
public sub ReadReport()
  #if (USB_SERVICE)     // intr method (from 46J50 port)
    // djb 06-09-2010
    while ((mHIDRxIsBusy <> 0) and Attached)
    end while
    HIDRxReport(@RXReport, HID_REPORT_OUT_SIZE)
  #else                 // original code (intr or polled)
    DisableISR()
    while ((mHIDRxIsBusy <> 0) and Attached)
        _service()
    end while
    HIDRxReport(@RXReport, HID_REPORT_OUT_SIZE)
    USBService()
    EnableISR()
  #endif
end sub

public sub WriteReport()
  #if (USB_SERVICE)     // intr method (from 46J50 port)
    // djb 06-09-2010
    while ((mHIDTxIsBusy <> 0) and Attached)
    end while
    HIDTxReport(@TXReport, HID_REPORT_IN_SIZE)
  #else                 // original code (intr or polled)
    DisableISR()
    WaitForTX()
    HIDTxReport(@TXReport, HID_REPORT_IN_SIZE)
    USBService()
    EnableISR()
  #endif
end sub
#endif      // USB_EXTENDED_RAM_SIZE

//
//**************************************************************************
//Name    : ReadArray (OVERLOADED)                                         *
//Purpose : Read an array of bytes - pCount is an optional number of bytes *
//        : to read. If pCount is not given, the size of the array is used *
//**************************************************************************
// note: this routine has been overloaded and split into two to add support
// for passing an address and count (for ReadReport() without ExtendedRam)
//
// read 'count' bytes into 'buf_addr' ram space. this works with any data type
public sub ReadArray(buf_addr as word, count as word)
    dim BytesRead as byte

    DisableISR()
    while ((count > 0) and Attached)
        // wait for data
        if (mHIDRxIsBusy = 0) then
            if (count > HID_REPORT_OUT_SIZE) then                // pCount > max report size
                BytesRead = HIDRxReport(buf_addr, HID_REPORT_OUT_SIZE)
                dec(count, BytesRead)
                inc(buf_addr, BytesRead)
            else                                               // pCount <= max report size
                BytesRead = HIDRxReport(buf_addr, count.byte0)
                dec(count.byte0, BytesRead)
                inc(buf_addr, BytesRead)
            endif
        endif
        _service()
    end while
    EnableISR()
end sub

public sub ReadArray(byref pBuffer() as byte, pCount as word = 0)
    if (pCount = 0) then
       pCount = bound(pBuffer) + 1
    endif
    ReadArray(addressof(pBuffer), pCount)
end sub

//
//**************************************************************************
//Name    : WriteArray (OVERLOADED)                                        *
//Purpose : Write an array of bytes - pCount is an optional number of      *
//        : to write. If pCount is not given, the array size is used       *
//        : Bytes are transfered in HIDReportBytesIn sized packets - 8, 16 *
//        : 32 or 64 bytes. If you pass an array which is only dimensioned *
//        : 4 bytes (for example, DIM Array(4) AS BYTE  then a             *
//        : 'HIDReportBytesIn' sized data packet is still transferred.     *
//        : This isn't really a problem, as the host application should    *
//        : know what to do with the data anyway.                          *
//**************************************************************************
// note: this routine has been overloaded and split into two to add support
// for passing an address and count (for WriteReport() without ExtendedRam)
//
// write 'len' bytes from 'buf_addr' ram space. this works with any data type
public sub WriteArray(buf_addr as word, len as word)
    DisableISR()
    while ((len > HID_REPORT_IN_SIZE) and Attached)
        WaitForTX()
        HIDTxReport(buf_addr, HID_REPORT_IN_SIZE)
        dec(len, HID_REPORT_IN_SIZE)
        inc(buf_addr, HID_REPORT_IN_SIZE)
        _service()
    end while

    if (len.byte0 > 0) then
        WaitForTX()
        HIDTxReport(buf_addr, HID_REPORT_IN_SIZE)
        _service()
    endif
    EnableISR()
end sub

public sub WriteArray(byref pBuffer() as byte, pLength as word = 0)
    if (pLength = 0) then
        pLength = bound(pBuffer) + 1
    endif
    WriteArray(addressof(pBuffer), pLength)
end sub

//
//**************************************************************************
//Name    : Initialize                                                     *
//Purpose : Initialize the USB module                                      *
//**************************************************************************
//
public sub Initialize()
    // setup PLL (see usbsfr.bas)
    SetUSBClock()

    mInitializeUSBDriver()

  #if (USB_SERVICE)
    #if (USB_SERVICE_PRIORITY = ipLow)
      IPR2Bits.USBIP = 0
    #else
      IPR2Bits.USBIP = 1
    #endif
    PIR2Bits.USBIF = 0      // clear USBIF
    PIE2Bits.USBIE = 1      // enable USB interrupt

    _service()
    UIE = UIE_ENABLE_ALL    // unmask all USB interrupts
    UEIE = UEIE_ENABLE_ALL  // unmask all USB error interrupts
    enable(USB_intr_handler)
  #endif    // USB_SERVICE
end sub

//
//**************************************************************************
// initialize the module
//**************************************************************************
//
#if (USB_AUTO_INITIALIZE)
  Initialize()
#else
  #if (SHOW_USB_BUILD_MSGS)
    #warning "USB_AUTO_INITIALIZE false... driver NOT initialized at startup"
  #endif
#endif

