//
//---------------------------------------------------------------------------
// Name    : userio.bas
// Author  : Jerry Messina
// Date    : 4/17/2011
// Version : 1.0
// Notes   : user hardware definitions/routines
//         : customize to match your hardware
//---------------------------------------------------------------------------
//
module userio

// note: if you do not need the following routines, then just comment out
// the code but leave the routines themselves... they will be optimized
// away if empty 

//
//-----------------------------------------------------------------------------
// User hardware routines
//-----------------------------------------------------------------------------
//
// here we define two pins... a status led output and a bootselect input
// the bootselect input can be used to tell the loader that we wish to stay
// here and not run the application

//
//-------------------------------------------------------------
// USB whacker hardware
//-------------------------------------------------------------
//
#if defined(USB_WHACKER)
  #if (SHOW_BUILD_MSGS)
    #warning "USB_WHACKER selected"
  #endif
  // usb whacker hardware (18F2553)
  public dim YELLOW_LED  as PORTC.0     // LED2
  public dim RED_LED     as PORTC.1     // LED3

  #option _STATUS_LED = PORTC.0         // LED2 (yellow)
  #option _BOOTSELECT_IN = PORTC.2      // pushbutton switch S2

//
//-------------------------------------------------------------
// Microchip LPC (LowPinCount) 14K50 USB development board
//-------------------------------------------------------------
//
#elseif defined(LPC_DEVBD)
  #if (SHOW_BUILD_MSGS)
    #warning "LPC_DEVBD selected"
  #endif
  // Microchip LPC (LowPinCount) USB development board
  #option _STATUS_LED = PORTC.0         // LED D1 (green)
  #option _BOOTSELECT_IN = PORTA.3      // pushbutton switch S1 **REQUIRES MCLRE config bit OFF**

//
//-------------------------------------------------------------
// 46J50 eval board hardware (USB self-power)
//-------------------------------------------------------------
//
#elseif defined(EVAL_J50)
  #if (SHOW_BUILD_MSGS)
    #warning "EVAL_J50 selected"
  #endif
  // 46J50 board hardware (USB self-powered)
  #option _STATUS_LED = PORTC.0         // STAT LED
  #option _BOOTSELECT_IN = PORTA.0      // S1 **DEFAULTS TO ANALOG AN0**

  // the J50 eval board is self-powered
  // Vusb is connected to RA5 via resistive divider
  #option USB_BUS_SENSE_IO = true                                                'usbcfg.h
  #option USB_BUS_SENSE_PIN = PORTA.5
  dim USB_BUS_SENSE_IN as USB_BUS_SENSE_PIN.bits(USB_BUS_SENSE_PIN@)
#endif


dim STATUS_LED  as _STATUS_LED.bits(_STATUS_LED@)
dim BOOTSELECT_IN as _BOOTSELECT_IN.bits(_BOOTSELECT_IN@)

//
//-----------------------------------------------------------------------------
// UserInit is called at program startup
//-----------------------------------------------------------------------------
//
public inline sub UserInit()
    // << add code to setup any user-defined hardware needed by the bootloader >>
    // << comment out code if not used >>

    // set led pin as output, low
    low(STATUS_LED)

    // set bootselect pin as an input
    input(BOOTSELECT_IN)

    // you may have to do more than just set the pin as an input...
    // such as disable any ADC channels, etc
  #if defined(EVAL_J50)
    ANCON0.bits(0) = 1                  // RA0 is AN0... set to digital
    // also, the J50 board is self-powered, so we need to set the Vusb sense pin
    input(USB_BUS_SENSE_IN)
    ANCON0.bits(4) = 1                  // RA5 is AN4... set to digital
  #endif
end sub

//
//-----------------------------------------------------------------------------
// UserExit is called before exiting the loader and running the application
//-----------------------------------------------------------------------------
//
public inline sub UserExit()
    // << add code to restore any hardware registers to their power on state >>
    // << comment out code if not used >>

    input(STATUS_LED)
    input(BOOTSELECT_IN)

  #if defined(EVAL_J50)         // set pins back to analog
    ANCON0.bits(0) = 0
    ANCON0.bits(4) = 1
  #endif
end sub

//
//-----------------------------------------------------------------------------
// User LED controls (optional)
//-----------------------------------------------------------------------------
//
public inline sub UserLedOn()
    // << add code to turn on an led >>
    // << comment out code if not used >>
    STATUS_LED = 1
end sub

public inline sub UserLedOff()
    // << add code to turn off an led >>
    // << comment out code if not used >>
    STATUS_LED = 0
end sub

//
//-----------------------------------------------------------------------------
// UserBootselect - read user-defined IO pin or switch used for 
//                  bootloader control
//   returns TRUE - activate bootloader
//           FALSE - skip bootloader and run application
//-----------------------------------------------------------------------------
//
public inline function UserBootselect() as boolean
    // << add code to read switch or IO pin >>
    // << just return false if not used >>
    ' result = false
    
    if (BOOTSELECT_IN = 0) then     // read low-active IO pin
        result = true               // stay in bootloader
    else
        result = false              // run application
    endif
end function

//
//-----------------------------------------------------------------------------
// toggle led to verify clock frequency
// (normally unused... for initial verification purposes only)
//-----------------------------------------------------------------------------
//
public sub User_check_clock()
    while(true)
        UserLedOn()
        delayus(500)
        UserLedOff()
        delayus(500)
    end while
end sub


end module
