//
//***************************************************************************
// Name    : USBSystem.Bas                                                  *
// Author  : David John Barker                                              *
// Notice  : Copyright (c) 2007 Mecanique                                   *
//         : All Rights Reserved                                            *
// Date    : 06/01/2007                                                     *
//                                                                          *
// Version : 1.4 Jerry Messina                                              *
//         : add minor code optimizations (reduces code size ~256 bytes)    *
//         :   (#define OPTIMIZE_USBSYSTEM)                                 *
//         : add use of local variable SetupPkt_W_Value in hid routines     *
//         :   HandleControlOutReport(), HIDGetReportHandler(), and         *
//         :   HIDSetReportHandler(). this saves ~44 bytes of code assuming *
//         :   the local gets allocated into bank 0 (this is useful for the *
//         :   hidbootloader)                                               *
//         : incorporate V1.3A Mchip Fritz Schlunder 04/14/09 fix           *
//         :   Improved SETUP response for hosts with aggressive control    *
//         :   transfer timing (#define INSTALL_MCHIP_USB_V13A_FIX)         *
//         :   note: this adds 20-30 bytes of code                          *
//         : add UCONBits.SUSPND = 0 to USBModuleDisable(), following the   *
//         :   datasheet recommendations                                    *
//         : change USBSoftDetach() to public                               *
//         : added code in USBStdGetDscHandler() to STALL for invalid       *
//         :   string descriptor request. Now passes Axelson's HidTest.exe  *
//         : add code in USBCtrlTrfSetupHandler() to see if CheckStdRequest *
//         :   handled the packet before calling HID/CDC Request            *
//         :                                                                *
//         : bug fixes to existing v1.3 code:                               *
//         : **BUG FIX 1.3-1: USBCheckStdRequest()                          *
//         :   change duplicate 'case CLR_FEATURE' to 'case SET_FEATURE'    *
//         : **BUG FIX 1.3-2: CDCTxService()                                *
//         :   change pointer increment                                     *
//         :   from:  inc(pCDCSrc.bRam, byte_to_send)                       *
//         :     to:  inc(pCDCSrc.bRom, byte_to_send)                       *
//         : **BUG FIX 1.3-3: USBCtrlTrfInHandler()                         *
//         :   add 'else' block to short_pkt_status test                    *
//         : **BUG FIX 1.3-4: USBDriverService()                            *
//         :   remove code at end of 'Task D' that should have been removed *
//         :   in a prior fix                                               *
//         : **BUG FIX 1.3-5: USBStdGetStatusHandler()                      *
//         :   move RemoteWakeup test outside of 'if (self_power)'          *
//                                                                          *
// Version : 1.3 PIC18F2450/4450 Data Sheet Errata                          *
//         : Microchip firmware bug fix - search for 'bug fix' in this file *
//         : 1.2 Added GetPID() to correctly mask BD_STAT PID               *
//         :     Added string index check in USBStdGetDscHandler()          *
//         :     Added 'Attached' flag                                      *
//         : 1.1 Fixed RAM allocation for smaller USB devices               *
//         : 1.0 Release                                                    *
// Notes   : This module is based on the Microchip USB C18 Firmware Version *
//         : 1.0 by Rawin Rojvanit. Whilst many of the routines could be    *
//         : optimised for the Swordfish compiler, it was felt that         *
//         : maintaining the code as close to possible to the original C    *
//         : version would make it much easier to make changes and update   *
//         : the firmware. In addition, all the documentation relating to   *
//         : the original version can be easily mapped to this module       *
//         : -------------------------------------------------------------- *
//         : IT IS STRONGLY advised you do not make changes to this module  *
//         : unless you are familiar with USB or the original Microchip     *
//         : firmware                                                       *
//***************************************************************************
//
module USBSystem

// import modules...
include "USBConfig.bas"     // configurations
include "USBSFR.bas"        // USB Special Function Registers (SFR)
include "usbmem.bas"        // usb memory layout
include "USBDefs.bas"       // constants, structures etc
include USB_DESCRIPTOR      // descriptor file

// install code optimizations (saves approx 256 bytes)
// note: not all optimizations are controlled by this setting
#define OPTIMIZE_USBSYSTEM

// install usb stack V1.3A modifications
// taken from Microchip USB HID bootloader, this fix is to improve SETUP response
// for hosts with aggressive control transfer timing
// enabling this will increase code size ~26 bytes
'#define INSTALL_MCHIP_USB_V13A_FIX

// general module variables...
dim                                                                      'usbmap.c
    usb_device_state as byte,           // Device States: DETACHED, ATTACHED, ...
    usb_stat as USB_DEVICE_STATUS,      // Global USB flags
    usb_active_cfg as byte,             // Value of current configuration
    usb_alt_intf(MAX_NUM_INT) as byte,  // Array to keep track of the current alternate setting for each interface ID
    usb_Attached as boolean             // device attached flag

// public...
public dim
    Attached as usb_Attached            // USB device attached flag

// verify max endpoints used
#if (MAX_EP_NUMBER > USB_ENDPOINT_REGISTERS)
  #error "MAX_EP_NUMBER exceeds device endpoint registers"
#endif

#define EP_SIZE = 4                     // endpoint size (in bytes)
#define USB_RAM_EP = USB_RAM_BASE       // the BDT endpoints are located at the start of USB RAM

// endpoint 0, OUT and IN
#if(0 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep0Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 0),
    ep0Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 1)
#endif

// endpoint 1, OUT and IN
#if(1 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep1Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 2),
    ep1Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 3)
#endif

// endpoint 2, OUT and IN
#if(2 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep2Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 4),
    ep2Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 5)
#endif

// endpoint 3, OUT and IN
#if(3 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep3Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 6),
    ep3Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 7)
#endif

// endpoint 4, OUT and IN
#if(4 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep4Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 8),
    ep4Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 9)
#endif

// endpoint 5, OUT and IN
#if(5 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep5Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 10),
    ep5Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 11)
#endif

// endpoint 6, OUT and IN
#if(6 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep6Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 12),
    ep6Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 13)
#endif

// endpoint 7, OUT and IN
#if(7 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep7Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 14),
    ep7Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 15)
#endif

// endpoint 8, OUT and IN
#if(8 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep8Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 16),
    ep8Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 17)
#endif

// endpoint 9, OUT and IN
#if(9 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep9Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 18),
    ep9Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 19)
#endif

// endpoint 10, OUT and IN
#if(10 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep10Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 20),
    ep10Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 21)
#endif

// endpoint 11, OUT and IN
#if(11 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep11Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 22),
    ep11Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 23)
#endif

// endpoint 12, OUT and IN
#if(12 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep12Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 24),
    ep12Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 25)
#endif

// endpoint 13, OUT and IN
#if(13 <= MAX_EP_NUMBER)
public dim                                                                      'usbmap.c
    ep13Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 26),
    ep13Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 27)
#endif

// endpoint 14, OUT and IN
#if(14 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep14Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 28),
    ep14Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 29)
#endif

// endpoint 15, OUT and IN
#if(15 <= MAX_EP_NUMBER)
dim                                                                             'usbmap.c
    ep15Bo as BDT absolute (USB_RAM_EP + EP_SIZE * 30),
    ep15Bi as BDT absolute (USB_RAM_EP + EP_SIZE * 31)
#endif

// located after endpoints...
#define USB_RAM_BUFF = USB_RAM_EP + EP_SIZE * (MAX_EP_NUMBER + 1) * 2
dim                                                                             'usbmap.c
    SetupPkt    as CTRL_TRF_SETUP   absolute USB_RAM_BUFF,
    CtrlTrfData as CTRL_TRF_DATA    absolute (USB_RAM_BUFF + USB_EP0_BUFF_SIZE)

// HID - located after endpoint buffers...
#if USB_USE_HID
#define USB_RAM_HID = USB_RAM_BUFF + (USB_EP0_BUFF_SIZE * 2)
dim                                                                             'usbmap.c
    hid_report_out(HID_INT_OUT_EP_SIZE)     as byte absolute USB_RAM_HID,
    hid_report_in(HID_INT_IN_EP_SIZE)       as byte absolute (USB_RAM_HID + HID_INT_OUT_EP_SIZE),
    hid_report_feature(USB_EP0_BUFF_SIZE)   as byte absolute (USB_RAM_HID + HID_INT_OUT_EP_SIZE + HID_INT_IN_EP_SIZE)
dim                                                                             'usbcfg.h
    HID_UEP     as UEP1,
    HID_BD_OUT  as ep1Bo,
    HID_BD_IN   as ep1Bi
#endif // USB_USE_HID

// CDC - located after endpoint buffers...
#if USB_USE_CDC
#define USB_RAM_CDC = USB_RAM_BUFF + (USB_EP0_BUFF_SIZE * 2)
dim                                                                             'usbmap.c
    cdc_notice(CDC_INT_EP_SIZE)         as byte absolute USB_RAM_CDC,
    cdc_data_rx(CDC_BULK_OUT_EP_SIZE)   as byte absolute (USB_RAM_CDC + CDC_INT_EP_SIZE),
    cdc_data_tx(CDC_BULK_IN_EP_SIZE)    as byte absolute (USB_RAM_CDC + CDC_INT_EP_SIZE + CDC_BULK_OUT_EP_SIZE)
dim
    CDC_COMM_UEP    as UEP2,                                                        'usbcfg.h
    CDC_INT_BD_IN   as ep2Bi,
    CDC_DATA_UEP    as UEP3,
    CDC_BULK_BD_OUT as ep3Bo,
    CDC_BULK_BD_IN  as ep3Bi
#endif // USB_USE_CDC

dim                                                                             'usbctrltr.h and usbctrltr.c
    ctrl_trf_state as byte,            // control transfer state
    ctrl_trf_session_owner as byte,    // session owner
    pSrc as POINTER,                   // source pointer
    pDst as POINTER,                   // destination pointer
    wCount as word

// originally coded in ROM, we have to use RAM in Swordfish...
dim                                                                             'usbdsc.c
    USB_CD_Ptr(2) as word,  // {&cfg01,&cfg01};
  #if USB_HAVE_SERIAL_STRING
    USB_SD_Ptr(4) as word   // {&sd000,&sd001,&sd002,&sd003};
  #else
    USB_SD_Ptr(3) as word   // {&sd000,&sd001,&sd002};
  #endif

//
//**************************************************************************
//Name    : SetRAMPtr                                                      *
//Purpose : Helper routine for setting RAM                                 *
//**************************************************************************
//
inline sub SetRAMPtr(pAddress as FSR0, pValue as POSTINC0)
end sub

//
//**************************************************************************
//Name    : GetRAMPtr                                                      *
//Purpose : Helper routine for getting RAM byte                            *
//**************************************************************************
//
inline function GetRAMPtr(pAddress as FSR1) as POSTINC1
end function

//
//**************************************************************************
//Name    : ReadROM                                                        *
//Purpose : Helper routine for getting ROM byte                            *
//**************************************************************************
//
inline function ReadROM() as TABLAT
    asm
        TBLRD *+
    end asm
end function

//
//**************************************************************************
//Name    :GetROMPtr                                                       *
//Purpose : Helper routine for getting ROM byte                            *
//**************************************************************************
// NOTE: this function assumes that the ROM string is located in page 0,
//       and that the TBLPTRU register is currently set to $00
inline function GetROMPtr(pAddress as TABLEPTR) as TABLAT
    ReadROM()
end function

//
//**************************************************************************
//Name    : GetRequestType                                                 *
//Purpose : Helper routine to obtain USB request type                      *
//**************************************************************************
//
function GetRequestType() as byte
    // result = (SetupPkt.RequestType_f2 >> 5) and %00000011
    result = SetupPkt.RequestType_f2
    result = result >> 5
    result = result and %00000011
end function

//
//**************************************************************************
//Name    : GetRecipient                                                   *
//Purpose : Helper routine to get USB recipient                            *
//**************************************************************************
//
function GetRecipient() as byte
    // result = SetupPkt.Recipient_f5 and $1F
    result = SetupPkt.Recipient_f5
    result = result and $1F
end function

//
//**************************************************************************
//Name    : GetEPNum                                                       *
//Purpose : Helper routineto get USB EP number                             *
//**************************************************************************
//
function GetEPNum() as byte
    // result = SetupPkt.EPNum_f4 and $0F
    result = SetupPkt.EPNum_f4
    result = result and $0F
end function

//
//**************************************************************************
//Name    : GetPID (version 1.2)                                           *
//Purpose : Helper routineto get USB PID                                   *
//**************************************************************************
//
function GetPID() as byte
    // result = (ep0Bo.Stat.PID >> 2) and $0F
    result = ep0Bo.Stat.PID
    result = result >> 2
    result = result and $0F
end function

//
//**************************************************************************
//Name    : mInitializeUSBDriver                                           *
//Purpose : Configures the USB module                                      *
//        : This register determines: USB Speed, On-chip pull-up resistor  *
//        : selection, On-chip tranceiver selection, bus eye pattern       *
//        : generation mode, Ping-pong buffering mode selection            *
//**************************************************************************
//
public sub mInitializeUSBDriver()                                               'usbdrv.h
    UCFG = UCFG_VAL
    usb_device_state = DETACHED_STATE
    usb_stat._byte = $00
    usb_active_cfg = $00
end sub

//
//**************************************************************************
//Name    : mDisableAllEP(was mDisableEP1to15)                             *
//Purpose : This routine disables all endpoints except EP0. It should be   *
//        : called when the host sends RESET or SET_CONFIGURATION request  *
//**************************************************************************
//
sub mDisableAllEP()                                                         'usbdrv.h
    UEP1 = $00
    UEP2 = $00
    UEP3 = $00
    UEP4 = $00
    UEP5 = $00
    UEP6 = $00
    UEP7 = $00
  #if (USB_ENDPOINT_REGISTERS > 8)
    UEP8 = $00
    UEP9 = $00
    UEP10 = $00
    UEP11 = $00
    UEP12 = $00
    UEP13 = $00
    UEP14 = $00
    UEP15 = $00
  #endif
end sub

//
//**************************************************************************
//Name    : mUSBBufferReady                                                *
//Require : IN Endpoint: Buffer is loaded and ready to be sent.            *
//        : OUT Endpoint: Buffer is free to be written to by SIE.          *
//Input   : buffer_dsc: Root name of the buffer descriptor group. i.e.     *
//        : ep0Bo, ep1Bi, ... Declared IN usbmmap.c. Names can be          *
//        : remapped for readability, see examples in usbcfg.h             *
//Purpose : This macro should be called each time after:                   *
//        :   1. A non-EP0 IN endpoint buffer is populated with data       *
//        :   2. A non-EP0 OUT endpoint buffer is read                     *
//        : This macro turns the buffer ownership to SIE for servicing     *
//        : It also toggles the DTS bit for synchronization                *
//**************************************************************************
//
#if defined(OPTIMIZE_USBSYSTEM)
// this optimization saves 120 bytes
// note: the functionality is slightly different, as the original code changes
// the descriptor status byte three times. this should not cause any issues
// since we should already own the descriptor
//
sub mUSBBufferReady(byref buffer_dsc as BDT)                                    'usbdrv.h
    dim b as byte
    b = buffer_dsc.Stat._byte           // read indirect byte
    b = b and _DTSMASK                  // save only DTS bit
    b = b xor _DTSMASK                  // toggle DTS bit
    b = b or (_USIE or _DTSEN)          // turn ownership to SIE
    buffer_dsc.Stat._byte = b           // update descriptor
end sub
#else
// original version
sub mUSBBufferReady(byref buffer_dsc as BDT)                                    'usbdrv.h
    buffer_dsc.Stat._byte = buffer_dsc.Stat._byte and _DTSMASK        // Save only DTS bit
    buffer_dsc.Stat.DTS = not buffer_dsc.Stat.DTS                     // Toggle DTS bit
    buffer_dsc.Stat._byte = buffer_dsc.Stat._byte or _USIE or _DTSEN  // Turn ownership to SIE
end sub
#endif

//
//**************************************************************************
//Name    : mUSBCheckAdrPendingState                                       *
//Purpose : Specialized checking routine, it checks if the device is in    *
//        : the ADDRESS PENDING STATE and services it if it is.            *
//**************************************************************************
//
sub mUSBCheckAdrPendingState()                                                  'usb9.h
    if (usb_device_state = ADR_PENDING_STATE) then
        UADDR = SetupPkt.bDevADR
        if (UADDR > 0) then
            usb_device_state = ADDRESS_STATE
        else
            usb_device_state = DEFAULT_STATE
        endif
    endif
end sub

//  ============================ HID CLASS ============================
#if USB_USE_HID

dim hid_rpt_rx_len as byte

// make public for HID module...
public dim HIDRXBytesAvailable as HID_BD_OUT.Cnt

//
//**************************************************************************
//Name    : mUSBGetHIDDscAdr                                               *
//Purpose : Get HID descriptor address                                     *
//**************************************************************************
//
sub mUSBGetHIDDscAdr()                                                          'usbcfg.h
    if (usb_active_cfg = 1) then
        // get a pointer to @cfg01.hid_i00a00 (HID Class-Specific Descriptor)
        pSrc.bRom = USB_CD_Ptr(1)                                   // @cfg01
        inc(pSrc.bRom, sizeof(USB_CFG_DSC) + sizeof(USB_INTF_DSC))  // offset of hid_i00a00
    endif
end sub

//
//**************************************************************************
//Name    : mUSBGetHIDRptDscAdr                                            *
//Purpose : Get HID descriptor report address                              *
//**************************************************************************
//
sub mUSBGetHIDRptDscAdr()                                                       'usbcfg.h
    if (usb_active_cfg = 1) then
        pSrc.bRom = @hid_rpt01
    endif
end sub

//
//**************************************************************************
//Name    : mUSBGetHIDRptDscSize                                           *
//Purpose : Get HID report descriptor size                                 *
//**************************************************************************
//
sub mUSBGetHIDRptDscSize()                                                      'usbcfg.h
    if (usb_active_cfg = 1) then
        wCount = sizeof(hid_rpt01)
    endif
end sub

//
//**************************************************************************
//Name    : mHIDRxIsBusy                                                   *
//Purpose : This macro is used to check if HID OUT endpoint is busy (owned *
//        : by SIE) or not.                                                *
//**************************************************************************
//
public dim mHIDRxIsBusy as HID_BD_OUT.Stat.UOWN                                 'hid.h

//
//**************************************************************************
//Name    : mHIDTxIsBusy                                                   *
//Purpose : This macro is used to check if HID IN endpoint is busy (owned  *
//        : by SIE) or not.                                                *
//**************************************************************************
//
public dim mHIDTxIsBusy as HID_BD_IN.Stat.UOWN                                  'hid.h

//
//**************************************************************************
//Name    : HandleControlOutReport                                         *
//Purpose : Checks to see if an output or Feature report has arrived on    *
//        : the control pipe. if yes, extracts and uses the data.          *
//**************************************************************************
//
sub HandleControlOutReport()                                                    'user_generic_hid.c
    dim count as byte
    dim SetupPkt_W_Value as word        // using a local here can save code

    SetupPkt_W_Value = SetupPkt.W_Value

    // Find out if an Output or Feature report has arrived on the control pipe
    // Get the report type from the Setup packet
    select (SetupPkt_W_Value.byte1)
        case $02    // Output report
            // Get the report ID from the Setup packet
            select (SetupPkt_W_Value.byte0)
                case 0      // Report ID 0
                    // This example application copies the Output report data to hid_report_in
                    // (Assumes Input and Output reports are the same length)
                    // A "real" application would do something more useful with the data

                    // wCount holds the number of bytes read in the Data stage
                    // This example assumes the report fits in one transaction
                    count = 0
                    while (count < HID_OUTPUT_REPORT_BYTES)
                        hid_report_in(count) = hid_report_out(count)
                        inc(count)
                    end while
                    // The number of bytes in the report (from usbcfg.h)
            end select

        case $03    // Feature report
            // Get the report ID from the Setup packet
            select (SetupPkt_W_Value.byte0)
                case 0      // Report ID 0
                    // The Feature report data is in hid_report_feature
                    // This example application just sends the data back in the next
                    // Get_Report request for a Feature report

                    // wCount holds the number of bytes read in the Data stage
                    // This example assumes the report fits in one transaction
            end select
    end select
end sub

//
//**************************************************************************
//Name    : HIDGetReportHandler                                            *
//Purpose : Called on receiving a Setup packet with a GET_REPORT request.  *
//        : Extracts the report type and the report ID of the report the   *
//        : device will send in the Data stage of the transfer. Calls a    *
//        : function to provide the data.                                  *
//**************************************************************************
//
sub HIDGetReportHandler()                                                       'hid.c

    dim SetupPkt_W_Value as word        // using a local here can save code

    SetupPkt_W_Value = SetupPkt.W_Value

    // The report type is in the high byte of the setup packet's Value field
    // 3 = Feature; 1 = Input

    select (SetupPkt_W_Value.byte1)
        case 1          // Input report
            // Find out which report ID was specified
            // The report ID is in the low byte (LSB) of the Value field

            // This example supports Report ID 0
            select (SetupPkt_W_Value.byte0)
                case 0      // Report ID 0
                            // HID will handle the request
                    ctrl_trf_session_owner = MUID_HID

                    // Provide the data to send
                    pSrc.bRam = @hid_report_in
                    wCount.byte0 = HID_INPUT_REPORT_BYTES // The number of bytes in the report
                case 1      // Report ID 1
                            // Place similar code to handle Report ID 1 here
            end select

        case 3          // Feature report
            // The report ID is in the low byte of the Setup packet's Value field
            // This example supports Report ID 0
            select (SetupPkt_W_Value.byte0)
                case 0      // Report ID 0
                    // The HID class code will handle the request
                    ctrl_trf_session_owner = MUID_HID
                    // Provide the data to send
                    pSrc.bRam = @hid_report_feature
                    wCount.byte0 = HID_FEATURE_REPORT_BYTES // The number of bytes in the report
            end select
    end select

    // The data is in RAM
    usb_stat.ctrl_trf_mem = _RAM               // Set memory type
end sub

//
//**************************************************************************
//Name    : HIDSetReportHandler                                            *
//Purpose : Called on receiving a Setup packet with a SET_REPORT request.  *
//           : Extracts the report type and the report ID of the report the   *
//        : host will send IN the Data stage of the transfer. Places the   *
//        : report type and report ID IN out_report_status.                *
//**************************************************************************
//
sub HIDSetReportHandler()                                                       'hid.c

    dim SetupPkt_W_Value as word        // using a local here can save code

    SetupPkt_W_Value = SetupPkt.W_Value

    // The report type is in the high byte of the Setup packet's Value field
    // 3 = Feature; 2 = Output

    select (SetupPkt_W_Value.byte1)
        case 2      // Output report
            // The report ID is in the low byte of the Setup packet's Value field
            // This example supports Report ID 0
            select (SetupPkt_W_Value.byte0)
                case 0  // Report ID 0
                    // HID will handle the request
                    ctrl_trf_session_owner = MUID_HID
                    // Save the report type and report ID in out_report_status

                    //out_report_status = SetOutReportStatus
                    //  (MSB(SetupPkt.W_Value),
                    //  LSB(SetupPkt.W_Value));

                    // When the report arrives in the data stage, the data will be
                    // stored in hid_report_out (defined in usbmmap.c)

                    pDst.bRam = @hid_report_out
            end select

        case 3      // Feature report
            // Find out which report ID was specified
            // The report ID is in the low byte (LSB) of the Value field

            // This example supports Report ID 0
            select (SetupPkt_W_Value.byte0)
                case 0      // Report ID 0
                    // HID will handle the request
                    ctrl_trf_session_owner = MUID_HID
                    // Save the report type and report ID in out_report_status

                    //out_report_status = SetOutReportStatus
                    //  (MSB(SetupPkt.W_Value),
                    //  LSB(SetupPkt.W_Value));

                    // When the report arrives in the data stage, the data will be
                    // stored in hid_report_feature (defined in usbmmap.c)
                    pDst.bRam = @hid_report_feature
            end select
    end select
end sub

//
//**************************************************************************
//Name    : USBCheckHIDRequest                                             *
//Purpose : This routine checks the setup data packet to see if it knows   *
//        : how to handle it                                               *
//**************************************************************************
//
// variables -  used in USBCheckHIDRequest                                      'hid.c
dim idle_rate as byte
dim active_protocol as byte   // [0] Boot Protocol [1] Report Protocol

sub USBCheckHIDRequest()                                                        'hid.c
    dim SetupPkt_bRequest as byte        // using a local here can save code
    dim SetupPkt_bDscType as byte

    // The packet must be directed to an interface with the same ID as the device's HID interface
    if (GetRecipient() <> RCPT_INTF) then
        exit
    endif
    if (SetupPkt.bIntfID <> HID_INTF_ID) then
        exit
    endif

    SetupPkt_bRequest = SetupPkt.bRequest
    SetupPkt_bDscType = SetupPkt.bDscType

    // There are two standard requests that hid.c may support
    // 1 -  GET_DSC(DSC_HID,DSC_RPT,DSC_PHY)
    // 2 -  SET_DSC(DSC_HID,DSC_RPT,DSC_PHY)
    if (SetupPkt_bRequest = GET_DSC) then
        // Request for a descriptor
        select (SetupPkt_bDscType)
            // HID descriptor
            case DSC_HID
                ctrl_trf_session_owner = MUID_HID
                mUSBGetHIDDscAdr()        // See usbcfg.h
                wCount = sizeof(USB_HID_DSC)

            // Report descriptor
            case DSC_RPT
                ctrl_trf_session_owner = MUID_HID
                mUSBGetHIDRptDscAdr()     // See usbcfg.h
                mUSBGetHIDRptDscSize()    // See usbcfg.h

            // Physical descriptor (optional)
            case DSC_PHY
                // ctrl_trf_session_owner = MUID_HID
        end select
        usb_stat.ctrl_trf_mem = _ROM
    endif

    if (GetRequestType() <> CLASS) then
        exit
    endif

    // HID-specific requests
    select (SetupPkt_bRequest)
        // required for all HID tranfers
        case GET_REPORT
            HIDGetReportHandler()

         // optional - needed for control transfers only
        case SET_REPORT
            HIDSetReportHandler()

        // optional
        case GET_IDLE
            ctrl_trf_session_owner = MUID_HID
            pSrc.bRam = @idle_rate             // Set source
            usb_stat.ctrl_trf_mem = _RAM       // Set memory type
            wCount.byte0 = 1                   // Set data count

        // optional - except for keyboards using the boot protocol
        case SET_IDLE
            ctrl_trf_session_owner = MUID_HID
            idle_rate = SetupPkt.W_Value.byte1

        // required for HIDs that support the boot protocol
        case GET_PROTOCOL
            ctrl_trf_session_owner = MUID_HID
            pSrc.bRam = @active_protocol       // Set source
            usb_stat.ctrl_trf_mem = _RAM       // Set memory type
            wCount.byte0 = 1                   // Set data count

        // required for HIDs that support the boot protocol
        case SET_PROTOCOL
            ctrl_trf_session_owner = MUID_HID
            active_protocol = SetupPkt.W_Value.byte0
    end select
end sub

//
//**************************************************************************
//Name    : HIDInitEP                                                      *
//Purpose : HIDInitEP initializes HID endpoints, buffer descriptors,       *
//        : internal state-machine, and variables. It should be called     *
//        : after the USB host has sent OUT a SET_CONFIGURATION request.   *
//**************************************************************************
//
sub HIDInitEP()                                                                 'hid.c
    hid_rpt_rx_len =0

    HID_UEP = EP_OUT_IN or HSHK_EN              // Enable 2 data pipes

    HID_BD_OUT.Cnt = sizeof(hid_report_out)     // Set buffer size
    HID_BD_OUT.ADR = @hid_report_out            // Set buffer address
    HID_BD_OUT.Stat._byte = _USIE or _DAT0 or _DTSEN    // Set status

    //
    //  Do not have to init Cnt of IN pipes here
    //  Reason:  Number of bytes to send to the host varies from one transaction to
    //           another. Cnt should equal the exact number of bytes to transmit for
    //           a given IN transaction
    //           This number of bytes will only be known right before the data is sent
    //
    HID_BD_IN.ADR = @hid_report_in              // Set buffer address
    HID_BD_IN.Stat._byte = _UCPU or _DAT1       // Set status
end sub

//
//**************************************************************************
//Name    : HIDTxReport                                                    *
//Require : mHIDTxIsBusy() must return false. Value of 'len' must be equal *
//        : to or smaller than HID_INT_IN_EP_SIZE for an interrupt         *
//        : endpoint, the largest buffer size is 64 bytes.                 *
//Input:  : buffer  : POINTER to the starting location of data bytes       *
//        : len     : Number of bytes to be transferred                    *
//Purpose : Use this macro to transfer data located in data memory.        *
//        : NOTE : mHIDTxIsBusy() must return false before user can call   *
//        : this function. Unexpected behavior will occur if this function *
//        : when mHIDTxIsBusy                                              *
//**************************************************************************
//
public sub HIDTxReport(Buffer as FSR1, len as byte)                            'hid.c
    dim i as byte

    if (len > HID_INT_IN_EP_SIZE) then
        len = HID_INT_IN_EP_SIZE
    endif

    // Copy data from user's buffer to dual-ram buffer
    FSR0 = @hid_report_in
    i = 0
    while (i < len)
        POSTINC0 = POSTINC1
        inc(i)
    end while

    HID_BD_IN.Cnt = len
    mUSBBufferReady(HID_BD_IN)
end sub

//
//**************************************************************************
//Name    : HIDRxReport                                                    *
//Require : Value of input argument 'len' should be smaller than the       *
//        : maximum endpoint size responsible for receiving report data    *
//        : from USB host for HID CLASS. Input argument 'buffer' should    *
//        : point to a buffer area that is bigger or equal to the size     *
//        : specified by 'len'.                                            *
//Input   : buffer  : POINTER to where received bytes are to be stored     *
//        : len     : The number of bytes expected.                        *
//Output  : The number of bytes copied to buffer. Publicly accessible      *
//        : variable hid_rpt_rx_len is updated with the number of bytes    *
//        : copied to buffer. Once HIDRxReport is called, subsequent       *
//        : retrieval of hid_rpt_rx_len can be done by calling macro       *
//        : mHIDGetRptRxLength().                                          *
//Purpose:: HIDRxReport copies a string of bytes received through          *
//        : USB HID OUT endpoint to a user's specified location. It is a   *
//        : non-blocking function. It does not wait for data if there is   *
//        : no data available. Instead it returns '0' to notify the        *
//        : caller that there is no data available. If the actual number   *
//        : of bytes received is larger than the number of bytes expected  *
//        : (len), only the expected number of bytes specified will be     *
//        : copied to buffer. If the actual number of bytes received is    *
//        : smaller than the number of bytes expected (len), only the      *
//        : actual number of bytes received will be copied to buffer.      *
//**************************************************************************
//
public function HIDRxReport(Buffer as FSR0, len as byte) as byte               'hid.c
    hid_rpt_rx_len = 0

    if (mHIDRxIsBusy = 0) then
        // Adjust the expected number of bytes to equal the actual number of bytes received
        if (len > HID_BD_OUT.Cnt) then
            len = HID_BD_OUT.Cnt
        endif

        // Copy data from dual-ram buffer to user's buffer
        FSR1 = @hid_report_out
        hid_rpt_rx_len = 0
        while (hid_rpt_rx_len < len)
            POSTINC0 = POSTINC1
            inc(hid_rpt_rx_len)
        end while

        // Prepare dual-ram buffer for next OUT transaction
        HID_BD_OUT.Cnt = sizeof(hid_report_out)
        mUSBBufferReady(HID_BD_OUT)
    endif
    result = hid_rpt_rx_len
end function
#endif // USB_USE_HID
//  ========================== END HID CLASS ==========================

//  ============================ CDC CLASS ============================
#if USB_USE_CDC
dim                                                                             'cdc.c
    cdc_trf_state as byte,            // States are defined cdc.h
    pCDCSrc as POINTER,               // Dedicated source pointer
    pCDCDst as POINTER,               // Dedicated destination pointer
    cdc_tx_len as word,               // total tx length
    cdc_mem_type as byte,             // _ROM, _RAM
    line_coding as CDC_LINE_CODING,   // Buffer to store line coding information
    control_signal_bitmap as CDC_CONTROL_SIGNAL_BITMAP

// make public for CDC module...
public dim
    CDCControlSignal as control_signal_bitmap,
    CDCDataRX as cdc_data_rx,
    CDCDataTX as cdc_data_tx,
    CDCRXBytesAvailable as CDC_BULK_BD_OUT.Cnt

// SEND_ENCAPSULATED_COMMAND and GET_ENCAPSULATED_RESPONSE are required
// requests according to the CDC specification
// However, it is not really being used here, therefore a dummy buffer is
// used for conformance
const dummy_length = $08
dim dummy_encapsulated_cmd_response(dummy_length) as byte

//
//**************************************************************************
//Name    : mUSBUSARTIsTxTrfReady                                          *
//Purpose : This macro is used to check if the CDC CLASS is ready to send  *
//        : more data.                                                     *
//**************************************************************************
//
public function mUSBUSARTIsTxTrfReady() as boolean                              'cdc.h
    result = (cdc_trf_state = CDC_TX_READY)
end function

//
//**************************************************************************
//Name    : mCDCUsartRxIsBusy                                              *
//Purpose : This macro is used to check if CDC bulk OUT endpoint is busy   *
//        : (owned by SIE) or not.                                         *
//**************************************************************************
//
public dim mCDCUsartRxIsBusy as CDC_BULK_BD_OUT.Stat.UOWN                       'cdc.h

//
//**************************************************************************
//Name    : mCDCUsartTxIsBusy                                              *
//Purpose : This macro is used to check if CDC bulk IN endpoint is busy    *
//        : (owned by SIE) or not.                                         *
//**************************************************************************
//
public dim mCDCUsartTxIsBusy as CDC_BULK_BD_IN.Stat.UOWN                        'cdc.h

//
//**************************************************************************
//Name    : mUSBUSARTTxRam                                                 *
//Require : cdc_trf_state must be IN the CDC_TX_READY state.               *
//Purpose : Use this macro to transfer data located IN data memory. Use    *
//        : this macro when:                                               *
//        :    1. Data stream is not null-terminated                       *
//        :    2. Transfer length is known                                 *
//        : Remember: cdc_trf_state must == CDC_TX_READY. The actual       *
//        : transfer is handled by CDCTxService().                         *
//**************************************************************************
//
public sub mUSBUSARTTxRam(pData as word, len as word)                            'cdc.h
    pCDCSrc.bRam = pData
    cdc_tx_len = len
    cdc_mem_type = _RAM
    cdc_trf_state = CDC_TX_BUSY
end sub

//
//**************************************************************************
//Name    : mUSBUSARTTxRom                                                 *
//Require : cdc_trf_state must be IN the CDC_TX_READY state.               *
//Purpose : Use this macro to transfer data located IN data memory. Use    *
//        : this macro when:                                               *
//        :    1. Data stream is not null-terminated                       *
//        :    2. Transfer length is known                                 *
//        : Remember: cdc_trf_state must == CDC_TX_READY. The actual       *
//        : transfer is handled by CDCTxService().                         *
//**************************************************************************
//
public sub mUSBUSARTTxRom(pData as word, len as word)                            'cdc.h
    pCDCSrc.bRom = pData
    cdc_tx_len = len
    cdc_mem_type = _ROM
    cdc_trf_state = CDC_TX_BUSY
end sub

//
//**************************************************************************
//Name    : USBCheckCDCRequest                                             *
//Purpose : This routine checks the setup data packet to see if it knows   *
//        : how to use it                                                  *
//**************************************************************************
//
sub USBCheckCDCRequest()                                                        'cdc.c

    // If request recipient is not an interface then return
    if (GetRecipient() <> RCPT_INTF) then
        exit
    endif

    // If request type is not class-specific then return
    if (GetRequestType() <> CLASS) then
        exit
    endif

    // Interface ID must match interface numbers associated with
    // CDC class, else return
    if ((SetupPkt.bIntfID <> CDC_COMM_INTF_ID) and (SetupPkt.bIntfID <> CDC_DATA_INTF_ID)) then
        exit
    endif

    select (SetupPkt.bRequest)
        case SEND_ENCAPSULATED_COMMAND
            ctrl_trf_session_owner = MUID_CDC
            pSrc.bRam = @dummy_encapsulated_cmd_response
            usb_stat.ctrl_trf_mem = _RAM
            wCount.byte0 = dummy_length

        case GET_ENCAPSULATED_RESPONSE
            ctrl_trf_session_owner = MUID_CDC
            // Populate dummy_encapsulated_cmd_response first
            pDst.bRam = @dummy_encapsulated_cmd_response

        //case SET_COMM_FEATURE                   // Optional
        //case GET_COMM_FEATURE                   // Optional
        //case CLEAR_COMM_FEATURE                 // Optional

        case SET_LINE_CODING
            ctrl_trf_session_owner = MUID_CDC
            pDst.bRam = @LINE_CODING    // Set destination

        case GET_LINE_CODING
            ctrl_trf_session_owner = MUID_CDC
            pSrc.bRam = @LINE_CODING           // Set source
            usb_stat.ctrl_trf_mem = _RAM       // Set memory type
            wCount.byte0 = LINE_CODING_LENGTH  // Set data count

        case SET_CONTROL_LINE_STATE
            ctrl_trf_session_owner = MUID_CDC
            control_signal_bitmap._byte = SetupPkt.W_Value.byte0

        //case SEND_BREAK                         // Optional
    end select
end sub

//
//**************************************************************************
//Name    : CDCInitEP                                                      *
//Purpose : CDCInitEP initializes CDC endpoints, buffer descriptors,       *
//        : internal state-machine, and variables. It should be called     *
//        : after the USB host has sent OUT a SET_CONFIGURATION request.   *
//**************************************************************************
//
sub CDCInitEP()                                                                 'cdc.c
    // abstract line coding information
    LINE_CODING.dwDTERate = 115200      // baud rate
    LINE_CODING.bCharFormat = $00       // 1 stop bit
    LINE_CODING.bParityType = $00       // None
    LINE_CODING.bDataBits = $08         // 5,6,7,8, or 16

    cdc_trf_state = CDC_TX_READY

    CDC_COMM_UEP = EP_IN or HSHK_EN               // Enable 1 Comm pipe
    CDC_DATA_UEP = EP_OUT_IN or HSHK_EN           // Enable 2 data pipes
    //
    // Do not have to init Cnt of IN pipes here... the number of bytes to
    // send to the host varies from one transaction to another. Cnt should
    // equal the exact number of bytes to transmit for a given IN transaction.
    // This number of bytes will only be known right before the data is sent
    //
    CDC_INT_BD_IN.ADR = @cdc_notice               // Set buffer address
    CDC_INT_BD_IN.Stat._byte = _UCPU or _DAT1     // Set status

    CDC_BULK_BD_OUT.Cnt = sizeof(cdc_data_rx)     // Set buffer size
    CDC_BULK_BD_OUT.ADR = @cdc_data_rx            // Set buffer address
    CDC_BULK_BD_OUT.Stat._byte = _USIE or _DAT0 or _DTSEN // Set status

    CDC_BULK_BD_IN.ADR = @cdc_data_tx             // Set buffer address
    CDC_BULK_BD_IN.Stat._byte = _UCPU or _DAT1    // Set status
end sub

//
//**************************************************************************
//Name    : CDCRXPrepareForNext                                            *
//Purpose : Prepare for next CDC transmission                              *
//**************************************************************************
//
public sub CDCRXPrepareForNext()
    CDC_BULK_BD_OUT.Cnt = sizeof(cdc_data_rx)
    mUSBBufferReady(CDC_BULK_BD_OUT)
end sub

//
//**************************************************************************
//Name    : CDCTxService                                                   *
//Purpose : CDCTxService handles device-to-host transaction(s). This       *
//        : function should be called once per Main program loop.          *
//**************************************************************************
//
public sub CDCTxService()                                                       'cdc.c
    dim byte_to_send as byte

    if (mCDCUsartTxIsBusy <> 0) then
        exit
    endif

    // Completing stage is necessary while [ mCDCUSartTxIsBusy()==1 ]
    // By having this stage, user can always check cdc_trf_state,
    // and not having to call mCDCUsartTxIsBusy() directly
    if (cdc_trf_state = CDC_TX_COMPLETING) then
        cdc_trf_state = CDC_TX_READY
    endif

    // If CDC_TX_READY state, nothing to do, just return
    if (cdc_trf_state = CDC_TX_READY) then
        exit
    endif

    // If CDC_TX_BUSY_ZLP state, send zero length packet
    if (cdc_trf_state = CDC_TX_BUSY_ZLP) then
        CDC_BULK_BD_IN.Cnt = 0
        cdc_trf_state = CDC_TX_COMPLETING
    elseif (cdc_trf_state = CDC_TX_BUSY) then
        // First, have to figure out how many byte of data to send
        if (cdc_tx_len > sizeof(cdc_data_tx)) then
            byte_to_send = sizeof(cdc_data_tx)
        else
            byte_to_send = cdc_tx_len
        endif

        // Next, load the number of bytes to send to Cnt in buffer descriptor
        CDC_BULK_BD_IN.Cnt = byte_to_send

        // Subtract the number of bytes just about to be sent from the total
        cdc_tx_len = cdc_tx_len - byte_to_send

        pCDCDst.bRam = @cdc_data_tx     // Set destination pointer
        if (cdc_mem_type = _ROM) then     // Determine type of memory source
            TABLEPTR = pCDCSrc.bRom
            FSR0 = pCDCDst.bRam
            inc(pCDCDst.bRam, byte_to_send)
            // **BUG FIX 1.3-2 fix pointer increment - was 'inc(pCDCSrc.bRam, byte_to_send)'
            inc(pCDCSrc.bRom, byte_to_send)
            while (byte_to_send > 0)
                POSTINC0 = ReadROM()
                dec(byte_to_send)
            end while
        else   // _RAM - optimised...
            FSR0 = pCDCDst.bRam
            FSR1 = pCDCSrc.bRam
            inc(pCDCDst.bRam, byte_to_send)
            inc(pCDCSrc.bRam, byte_to_send)
            while (byte_to_send > 0)
                POSTINC0 = POSTINC1
                dec(byte_to_send)
            end while
        endif

        // Lastly, determine if a zero length packet state is necessary
        // See explanation in USB Specification 2.0: Section 5.8.3
        if (cdc_tx_len = 0) then
            if (CDC_BULK_BD_IN.Cnt = sizeof(cdc_data_tx)) then
                cdc_trf_state = CDC_TX_BUSY_ZLP
            else
                cdc_trf_state = CDC_TX_COMPLETING
            endif
        endif
    endif

    // Both CDC_TX_BUSY and CDC_TX_BUSY_ZLP states use the following macro
    mUSBBufferReady(CDC_BULK_BD_IN)
end sub
//
//  ========================== END CDC CLASS ==========================
//
#endif //USB_USE_CDC

//
//**************************************************************************
//Name    : USBStdGetDscHandler                                            *
//Purpose : This routine handles the STANDARD GET_DESCRIPTOR request. It   *
//        : utilizes tables dynamically looks up descriptor size. This     *
//        : routine should never have to be modified if the tables in      *
//        : usbdsc.c are declared correctly.                               *
//**************************************************************************
//
sub USBStdGetDscHandler()                                                       'usb9.c
    if (SetupPkt.bmRequestType = $80) then
        select (SetupPkt.bDscType)
            case DSC_DEV
                ctrl_trf_session_owner = MUID_USB9
                pSrc.bRom = @device_dsc         // djb USB_DD_Ptr  // @device_dsc
                wCount = sizeof(device_dsc)     // djb USB_DD_Size    // sizeof(device_dsc)
            case DSC_CFG
                ctrl_trf_session_owner = MUID_USB9
                pSrc.bRom = USB_CD_Ptr(SetupPkt.bDscIndex)
                wCount = GetROMPtr(pSrc.wRom + 2)    // Set data count (incr ptr to word, so +2)
            case DSC_STR
                ctrl_trf_session_owner = MUID_USB9
                // ensure we have a valid string index - version 1.2...
                if (SetupPkt.bDscIndex > bound(USB_SD_Ptr)) then
                    ctrl_trf_session_owner = MUID_NULL      // added V1.4 ... generate STALL
                    wCount = 0
                else
                    pSrc.bRom = USB_SD_Ptr(SetupPkt.bDscIndex)
                    wCount = GetROMPtr(pSrc.bRom)              // Set data count
                endif
        end select
        usb_stat.ctrl_trf_mem = _ROM                     // Set memory type
    endif
end sub

//
//**************************************************************************
//Name    : USBStdSetCfgHandler                                            *
//Purpose : This routine first disables all endpoints by clearing UEP      *
//        : registers. It then configures (initializes) endpoints specified*
//        : in the modifiable section.                                     *
//**************************************************************************
//
sub USBStdSetCfgHandler()                                                       'usb9.c
    ctrl_trf_session_owner = MUID_USB9
    mDisableAllEP()
    clear(usb_alt_intf)
    usb_active_cfg = SetupPkt.bCfgValue
    if (SetupPkt.bCfgValue = 0) then
        usb_device_state = ADDRESS_STATE
    else
        usb_device_state = CONFIGURED_STATE

        // ****modifiable section****
        #if USB_USE_HID
        HIDInitEP()
        #endif

        #if USB_USE_CDC
        CDCInitEP()
        #endif
        // ****end modifiable section****
    endif
end sub

//
//**************************************************************************
//Name    : USBStdGetStatusHandler                                         *
//Purpose : This routine handles the STANDARD GET_STATUS request           *
//**************************************************************************
//
// V1.4 optimization
// this function calculates the address:  @ep0Bo + (GetEPNum() * 8) + (SetupPkt.EPDir * 4)
// saves 124 bytes
function get_EP_addr() as word
    dim b as byte

    b = GetEPNum() * 8      // 0-15 * 8
    if (SetupPkt.EPDir = 1) then
        b = b + 4
    endif
    result = @ep0Bo
    result = result + b
end function

sub USBStdGetStatusHandler()                                                    'usb9.c
    CtrlTrfData._byte0 = 0                         // Initialize content
    CtrlTrfData._byte1 = 0

    select (GetRecipient())
        case RCPT_DEV
            ctrl_trf_session_owner = MUID_USB9
            // _byte0: bit0: Self-Powered Status [0] Bus-Powered [1] Self-Powered
            //         bit1: RemoteWakeup        [0] Disabled    [1] Enabled
            //
            //********************************************************************
            //Bug Fix: January 4, 2008
            //********************************************************************
            //Removed an extra '0' from the code:
            //CtrlTrfData._byte0|=0b00000001;
            //********************************************************************
            //
            if (self_power = 1) then                     // self_power defined in io_cfg.h
                CtrlTrfData._byte0 = CtrlTrfData._byte0 or %00000001    // Set bit0
            endif
            // **BUG FIX 1.3-5: move following code from inside previous 'if'
            if (usb_stat.RemoteWakeup = 1) then          // usb_stat defined in usbmmap.c
                CtrlTrfData._byte0 = CtrlTrfData._byte0 or $00000010     // Set bit1
            endif
        case RCPT_INTF
            ctrl_trf_session_owner = MUID_USB9                  // No data to update
        case RCPT_EP
            ctrl_trf_session_owner = MUID_USB9
            // _byte0: bit0: Halt Status [0] Not Halted [1] Halted
            pDst.bRam = get_EP_addr()                           // V1.4- added optimization
            if ((GetRAMPtr(pDst.bRam) and _BSTALL) <> 0) then   // Use _BSTALL as a bit mask
                CtrlTrfData._byte0 = $01                        // Set bit0
            endif
    end select

    if (ctrl_trf_session_owner = MUID_USB9) then
        pSrc.bRam = @CtrlTrfData            // Set Source
        usb_stat.ctrl_trf_mem = _RAM        // Set memory type
        wCount.byte0 = 2                    // Set data count
    endif
end sub

//
//**************************************************************************
//Name    : USBStdFeatureReqHandler                                        *
//Purpose : This routine handles the STANDARD SET & clear FEATURES requests*
//**************************************************************************
//
sub USBStdFeatureReqHandler()                                                   'usb9.c
    if ((SetupPkt.bFeature = DEVICE_REMOTE_WAKEUP) and (GetRecipient() = RCPT_DEV)) then
        ctrl_trf_session_owner = MUID_USB9
        if (SetupPkt.bRequest = SET_FEATURE) then
            usb_stat.RemoteWakeup = 1
        else
            usb_stat.RemoteWakeup = 0
        endif
    endif

    if ((SetupPkt.bFeature = ENDPOINT_HALT) and (GetRecipient() = RCPT_EP) and (GetEPNum() <> 0)) then
        ctrl_trf_session_owner = MUID_USB9
        pDst.bRam = get_EP_addr()       // V1.4- added optimization

        if (SetupPkt.bRequest = SET_FEATURE) then
            SetRAMPtr(pDst.bRam, _USIE or _BSTALL)
        else
            //
            //********************************************************************
            //Bug Fix: January 4, 2008
            //********************************************************************
            //Added '|DAT1' to force a halted IN endpoint to reinitialize to send
            //out DAT0 on the next transaction.
            //********************************************************************
            //
            if (SetupPkt.EPDir = 1) then        // IN
                SetRAMPtr(pDst.bRam, _UCPU or _DAT1)
            else
                SetRAMPtr(pDst.bRam, _USIE or _DAT0 or _DTSEN)
            endif
        endif
    endif
end sub

//
//**************************************************************************
//Name    : USBCheckStdRequest                                             *
//Purpose : This routine checks the setup data packet to see if it knows   *
//        : how to handle it                                               *
//**************************************************************************
//
sub USBCheckStdRequest()                                                        'usb9.c
    if (GetRequestType() <> STANDARD) then
        exit
    endif

    select (SetupPkt.bRequest)
        case SET_ADR
            ctrl_trf_session_owner = MUID_USB9
            usb_device_state = ADR_PENDING_STATE       // Update state only
            // See USBCtrlTrfInHandler() in usbctrltrf.c for the next step
        case GET_DSC
            USBStdGetDscHandler()
        case SET_CFG
            USBStdSetCfgHandler()
        case GET_CFG
            ctrl_trf_session_owner = MUID_USB9
            pSrc.bRam = @usb_active_cfg                 // Set Source
            usb_stat.ctrl_trf_mem = _RAM                // Set memory type
            wCount.byte0 = 1                            // Set data count
        case GET_STATUS
            USBStdGetStatusHandler()
        case CLR_FEATURE, SET_FEATURE    // **BUG FIX 1.3-1 remove duplicate case
            USBStdFeatureReqHandler()
        case GET_INTF
            ctrl_trf_session_owner = MUID_USB9
            pSrc.bRam = @usb_alt_intf + SetupPkt.bIntfID    // Set source
            usb_stat.ctrl_trf_mem = _RAM                    // Set memory type
            wCount.byte0 = 1                                // Set data count
        case SET_INTF
            ctrl_trf_session_owner = MUID_USB9
            usb_alt_intf(SetupPkt.bIntfID) = SetupPkt.bAltID
        case SET_DSC
        case SYNCH_FRAME
    end select
end sub

//
//********************************************************************
//Bug Fix: May 14, 2007 (#F7 - Partial 2/4)
//*********************************************************************
//For a control transfer read, if the host tries to read more data
//than what it has requested, the peripheral device should stall the
//extra IN transactions and the status stage. Typically, a host does
//not try to read more data than what it has requested. The original
//firmware did not handle this situation. Instead of stalling extra
//IN transactions, the device kept sending out zero length packets.
//
// This work around checks if a short IN packet has been sent or not.
// If it has, the IN endpoint will be set to install the next IN token.
// If not, then the original endpoint setup code will be executed.
//********************************************************************
//
dim short_pkt_status as byte              // Flag used by Control Transfer Read

//
//**************************************************************************
//Name    : USBCtrlTrfTxService                                            *
//Require : pSrc, wCount, and usb_stat.ctrl_trf_mem are setup properly.    *
//Purpose : This routine should be called from only two places. One from   *
//        : USBCtrlEPServiceComplete() and one from USBCtrlTrfInHandler(). *
//        : It takes care of managing a transfer over multiple USB         *
//        : transactions. This routine works with isochronous endpoint     *
//        : larger than 256 bytes and is shown here as an example of how to*
//        : deal with BC9 and BC8. IN reality, a control endpoint can never*
//        : be larger than 64 bytes.                                       *
//**************************************************************************
//
sub USBCtrlTrfTxService()                                                       'usbctrltr.c

    // This code executes in the Data stage of a Control Read transfer
    dim byte_to_send as word

    // First, have to figure out how many byte of data to send
    if (wCount < EP0_BUFF_SIZE) then
        byte_to_send = wCount
        //
        //********************************************************************
        //Bug Fix: May 14, 2007 (#F7 - Partial 3/4)
        //********************************************************************
        //For a control transfer read, if the host tries to read more data
        //than what it has requested, the peripheral device should stall the
        //extra IN transactions and the status stage. Typically, a host does
        //not try to read more data than what it has requested. The original
        //firmware did not handle this situation. Instead of stalling extra
        //IN transactions, the device kept sending out zero length packets.
        //
        //This work around updates the short_pkt_status flag to indicate
        //if a short packet has been sent or not. From this the state machine
        //can decide if it should stall future extra IN transactions or not
        //*******************************************************************
        //
        if (short_pkt_status = SHORT_PKT_NOT_USED) then
            short_pkt_status = SHORT_PKT_PENDING
        elseif (short_pkt_status = SHORT_PKT_PENDING) then
            short_pkt_status = SHORT_PKT_SENT
        endif
        //*******************************************************************
    else
        byte_to_send = EP0_BUFF_SIZE
    endif

    // Next, load the number of bytes to send to BC9..0 in buffer descriptor
    ep0Bi.Stat.BC9 = 0
    ep0Bi.Stat.BC8 = 0
    ep0Bi.Stat._byte = ep0Bi.Stat._byte or byte_to_send.byte1
    ep0Bi.Cnt = byte_to_send.byte0

    // Subtract the number of bytes just about to be sent from the total
    wCount = wCount - byte_to_send

    pDst.bRam = @CtrlTrfData                // Set destination pointer
    if (usb_stat.ctrl_trf_mem = _ROM) then  // Determine type of memory source
        TABLEPTR = pSrc.bRom
        FSR0 = pDst.bRam
        inc(pSrc.bRom, byte_to_send)
        while (byte_to_send > 0)
            POSTINC0 = ReadROM()
            dec(byte_to_send)
        end while
    else     // RAM
        FSR1 = pSrc.bRam
        FSR0 = pDst.bRam
        inc(pSrc.bRam, byte_to_send)
        while (byte_to_send > 0)
            POSTINC0 = POSTINC1
            dec(byte_to_send)
        end while
    endif
end sub

//
//**************************************************************************
//Name    : USBCtrlEPServiceComplete                                       *
//Purpose : This routine wrap up the ramaining tasks IN servicing a Setup  *
//        : Request. Its main task is to set the endpoint controls         *
//        : appropriately for a given situation. See code below.           *
//        : There are three main scenarios:                                *
//        :   a) There was no handler for the Request, IN this case a STALL*
//        :      should be sent OUT.                                       *
//        :   b) The host has requested a read control transfer, endpoints *
//        :      are required to be setup IN a specific way.               *
//        :   c) The host has requested a write control transfer, or a     *
//        :      control data stage is not required, endpoints are required*
//        :      to be setup IN a specific way.                            *
//        : Packet processing is resumed by clearing PKTDIS bit.           *
//**************************************************************************
//
sub USBCtrlEPServiceComplete()                                                  'usbctrltr.c
    // PKTDIS bit is set when a Setup Transaction is received
    // Clear before modifying ep0Bi.Stat or ep0Bo.Stat
    UCONBits.PKTDIS = 0

    if (ctrl_trf_session_owner = MUID_NULL) then
        // If no one knows how to service this request then stall
        // Must also prepare EP0 to receive the next SETUP transaction
        ep0Bo.Cnt = EP0_BUFF_SIZE
        ep0Bo.ADR = @SetupPkt

        ep0Bo.Stat._byte = _USIE or _BSTALL
        ep0Bi.Stat._byte = _USIE or _BSTALL
    else        // A module has claimed ownership of the control transfer session
        if (SetupPkt.DataDir = DEV_TO_HOST) then

            if (SetupPkt.wLength < wCount) then
                wCount = SetupPkt.wLength
            endif
            USBCtrlTrfTxService()
            ctrl_trf_state = CTRL_TRF_TX

            // Control Read:
            // <SETUP[0]><IN[1]><IN[0]>...<OUT[1]> | <SETUP[0]>
            // 1- Prepare OUT EP to respond to early termination
            // NOTE:
            // If something went wrong during the control transfer, the last status
            // stage may not be sent by the host. When this happens, two different
            // things could happen depending on the host...
            //   a) The host could send out a RESET
            //   b) The host could send out a new SETUP transaction without sending a RESET first
            // To properly handle case (b), the OUT EP must be setup to receive either
            // a zero length OUT transaction, or a new SETUP transaction
            //
            // Since the SETUP transaction requires the DTS bit to be DAT0 while the
            // zero length OUT status requires the DTS bit to be DAT1, the DTS bit check
            // by the hardware should be disabled. This way the SIE could accept either of
            // the two transactions
            //
            // Furthermore, the Cnt byte should be set to prepare for the SETUP data
            // (8-byte or more), and the buffer address should be pointed to SetupPkt
            ep0Bo.Cnt = EP0_BUFF_SIZE
            ep0Bo.ADR = @SetupPkt
            ep0Bo.Stat._byte = _USIE           // Note: DTSEN is 0!

            // 2- Prepare IN EP to transfer data
            //    Cnt should have been initialized by responsible request owner
            ep0Bi.ADR = @CtrlTrfData
            ep0Bi.Stat._byte = _USIE or _DAT1 or _DTSEN

        // (SetupPkt.DataDir == HOST_TO_DEV)
        else
            ctrl_trf_state = CTRL_TRF_RX
            // Control Write:
            // <SETUP[0]><OUT[1]><OUT[0]>...<IN[1]> | <SETUP[0]>
            // 1- Prepare IN EP to respond to early termination
            //    This is the same as a Zero Length Packet Response for control transfer
            //    without a data stage
            ep0Bi.Cnt = 0
            ep0Bi.Stat._byte = _USIE or _DAT1 or _DTSEN

            // Prepare OUT EP to receive data
            ep0Bo.Cnt = EP0_BUFF_SIZE
            ep0Bo.ADR = @CtrlTrfData
            ep0Bo.Stat._byte = _USIE or _DAT1 or _DTSEN
        endif
    endif
end sub

//
//**************************************************************************
//Name    : USBCtrlTrfSetupHandler                                         *
//Require : SetupPkt buffer is loaded with valid USB Setup Data            *
//Purpose : This routine is a task dispatcher and has 3 stages.            *
//        :      1. It initializes the control transfer state machine.     *
//        :      2. It calls on each of the module that may know how to    *
//        :         service the Setup Request from the host.               *
//        :         module Example: USB9, HID, CDC, MSD, ...               *
//        :         as new classes are added, ClassReqHandler table IN     *
//        :         usbdsc.c should be updated to call all available       *
//        :         CLASS handlers.                                        *
//        :      3. Once each of the modules has had a chance to check if  *
//        :         it is responsible for servicing the request, stage 3   *
//        :         then checks direction of the transfer to determine how *
//        :         to prepare EP0 for the control transfer.               *
//        :         Refer to USBCtrlEPServiceComplete() for more details.  *
//Note    : Microchip USB Firmware has three different states for the      *
//        : control transfer state machine:                                *
//        :      1. WAIT_SETUP                                             *
//        :      2. CTRL_TRF_TX                                            *
//        :      3. CTRL_TRF_RX                                            *
//        : A Control Transfer is composed of many USB transactions. When  *
//        : transferring data over multiple transactions, it is important  *
//        : to keep track of data source, data destination, and data count *
//        : These three parameters are stored IN pSrc,pDst, and wCount. A  *
//        : flag is used to note if the data source is from ROM or RAM.    *
//**************************************************************************
//
sub USBCtrlTrfSetupHandler()
    //
    //********************************************************************
    //Bug Fix: May 14, 2007 (#F7 - Partial 1/4)
    //*********************************************************************
    //For a control transfer read, if the host tries to read more data
    //than what it has requested, the peripheral device should stall the
    //extra IN transactions and the status stage. Typically, a host does
    //not try to read more data than what it has requested. The original
    //firmware did not handle this situation. Instead of stalling extra
    //IN transactions, the device kept sending out zero length packets.
    //
    //This work around forces the IN endpoint back to be owned by the
    //CPU after if it was stalled previously. The short_pkt_status flag
    //is also re-initialized.
    //********************************************************************
    //
    if (ep0Bi.Stat.UOWN <> 0) then
        ep0Bi.Stat._byte = _UCPU           // Compensate for after a STALL
    endif
    short_pkt_status = SHORT_PKT_NOT_USED

    // Stage 1
    ctrl_trf_state = WAIT_SETUP
    ctrl_trf_session_owner = MUID_NULL     // Set owner to NULL
    wCount = 0

    // Stage 2
    USBCheckStdRequest()                   // See system\usb9\usb9.c

    // ****Modifiable Section****
    // if not already handled, check device request (added in v1.4)
    if (ctrl_trf_session_owner = MUID_NULL) then
        #if USB_USE_HID
        USBCheckHIDRequest()
        #endif

        #if USB_USE_CDC
        USBCheckCDCRequest()
        #endif
    endif
    // ****End Modifiable Section****

    // Stage 3
    USBCtrlEPServiceComplete()
end sub

//
//**************************************************************************
//Name    : USBCtrlTrfRxService                                            *
//Require : pDst and wCount are setup properly. pSrc is always &CtrlTrfData*
//        : usb_stat.ctrl_trf_mem is always _RAM. wCount should be set to  *
//        : 0 at the start of each control transfer.                       *
//Purpose : This routine is only partially complete. Check for new version *
//        : of the firmware.                                               *
//**************************************************************************
//
sub USBCtrlTrfRxService()                                                       'usbctrltr.c
    // This code executes in the Data stage of a Control Write transfer
    dim byte_to_read as word

    byte_to_read.byte1 = $03 and ep0Bo.Stat._byte    // Filter out last 2 bits
    byte_to_read.byte0 = ep0Bo.Cnt

    // Accumulate total number of bytes read
    wCount = wCount + byte_to_read

    pSrc.bRam = @CtrlTrfData
    while (byte_to_read > 0)
        SetRAMPtr(pDst.bRam, GetRAMPtr(pSrc.bRam))
        inc(pDst.bRam)
        inc(pSrc.bRam)
        dec(byte_to_read)
    end while
end sub

//
//**************************************************************************
//Name    : USBPrepareForNextSetupTrf                                      *
//Purpose : The routine forces EP0 OUT to be ready for a new Setup         *
//        : transaction, and forces EP0 IN to be owned by CPU.             *
//**************************************************************************
//
#if defined(INSTALL_MCHIP_USB_V13A_FIX)
sub USBPrepareForNextSetupTrf()                                                 'usbctrltr.c
    ctrl_trf_state = WAIT_SETUP            // See usbctrltrf.h
    //
    //********************************************************************
    //Bug Fix: May 14, 2007 (#F3)
    //*********************************************************************
    //In the original firmware, if an IN token is sent by the host
    //before a SETUP token is sent, the firmware would respond with an ACK.
    //This is not a correct response, the firmware should have sent a STALL.
    //This is a minor non-compliance since a compliant host should not
    //send an IN before sending a SETUP token. The fix allows a SETUP
    //transaction to be accepted while stalling IN transactions.
    //
    //Although this fix is known, it is not implemented because it
    //breaks the #AF1 fix in USBCtrlEPServiceComplete().
    //Since #AF1 fix is more important, this fix, #F3 is commented out.
    //********************************************************************
    //
    ep0Bi.Stat._byte = _UCPU               // Should be removed
    //ep0Bi.Stat._byte = _USIE|_BSTALL;     // Should be added #F3
end sub
#else
sub USBPrepareForNextSetupTrf()                                                 'usbctrltr.c
    //
    //********************************************************************
    //Bug Fix: May 14, 2007
    //********************************************************************
    //Facts:
    //A Setup Packet should never be stalled. (USB 2.0 Section 8.5.3)
    //If a Setup PID is detected by the SIE, the DTSEN setting is ignored.
    //This causes a problem at the end of a control write transaction.
    //In USBCtrlEPServiceComplete(), during a control write (Host to Device),
    //the EP0_OUT is setup to write any data to the CtrlTrfData buffer.
    //If <SETUP[0]><IN[1]> is completed and USBCtrlTrfInHandler() is not
    //called before the next <SETUP[0]> is received, then the latest Setup
    //data will be written to the CtrlTrfData buffer instead of the SetupPkt
    //buffer.
    //
    //If USBCtrlTrfInHandler() was called before the latest <SETUP[0]> is
    //received, then there would be no problem,
    //because USBPrepareForNextSetupTrf() would have been called and updated
    //ep0Bo.ADR to point to the SetupPkt buffer.
    //
    //Work around:
    //Check for the problem as described above and copy the Setup data from
    //CtrlTrfData to SetupPkt.
    //********************************************************************
    //
    if ((ctrl_trf_state = CTRL_TRF_RX) and
        (UCONBits.PKTDIS = 1) and
        (ep0Bo.Cnt = sizeof(CTRL_TRF_SETUP)) and
        (ep0Bo.Stat.PID = SETUP_TOKEN) and
        (ep0Bo.Stat.UOWN = 0)) then

        ep0Bo.ADR = @SetupPkt

        // The Setup data was written to the CtrlTrfData buffer, must copy
        // it back to the SetupPkt buffer so that it can be processed correctly
        // by USBCtrlTrfSetupHandler().
        SetupPkt._byte = CtrlTrfData._byte
    // ******************************************************************
    else
        ctrl_trf_state = WAIT_SETUP            // See usbctrltrf.h
        ep0Bo.Cnt = EP0_BUFF_SIZE              // Defined in usbcfg.h
        ep0Bo.ADR = @SetupPkt

        //
        //********************************************************************
        //Bug Fix: May 14, 2007 (#F1)
        //********************************************************************
        //In the original firmware, if an OUT token is sent by the host
        //before a SETUP token is sent, the firmware would respond with an ACK.
        //This is not a correct response, the firmware should have sent a STALL.
        //This is a minor non-compliance since a compliant host should not
        //send an OUT before sending a SETUP token. The fix allows a SETUP
        //transaction to be accepted while stalling OUT transactions.
        //********************************************************************
        //
        //ep0Bo.Stat._byte = _USIE|_DAT0|_DTSEN;        // Removed
        ep0Bo.Stat._byte = _USIE or _DAT0 or _DTSEN or _BSTALL  // Added #F1

        //
        //********************************************************************
        //Bug Fix: May 14, 2007 (#F3)
        //********************************************************************
        //In the original firmware, if an IN token is sent by the host
        //before a SETUP token is sent, the firmware would respond with an ACK.
        //This is not a correct response, the firmware should have sent a STALL.
        //This is a minor non-compliance since a compliant host should not
        //send an IN before sending a SETUP token. The fix allows a SETUP
        //transaction to be accepted while stalling IN transactions.
        //
        //Although this fix is known, it is not implemented because it
        //breaks the #AF1 fix in USBCtrlEPServiceComplete().
        //Since #AF1 fix is more important, this fix, #F3 is commented out.
        //********************************************************************
        //
        ep0Bi.Stat._byte = _UCPU               // Should be removed
        //ep0Bi.Stat._byte = _USIE|_BSTALL;     // Should be added #F3
    endif

//
//    ctrl_trf_state = WAIT_SETUP            // See usbctrltrf.h
//    ep0Bo.Cnt = EP0BufferSize              // Defined in usbcfg.h
//    ep0Bo.ADR = @SetupPkt
//    ep0Bo.Stat._byte = _USIE Or _DAT0 Or _DTSEN  // EP0 buff dsc init, see usbmmap.h
//    ep0Bi.Stat._byte = _UCPU               // EP0 IN buffer initialization
//
end sub
#endif

//
//**************************************************************************
//Name    : USBCtrlTrfOutHandler                                           *
//Purpose : This routine handles an OUT transaction according to which     *
//        : control transfer state is currently active. Note that if the   *
//        : the control transfer was from host to device, the session owner*
//        : should be notified at the end of each OUT transaction to       *
//        : service the received data.                                     *
//**************************************************************************
//
sub USBCtrlTrfOutHandler()                                                      'usbctrltr.c
    // Direction of data packet: host to device
    if (ctrl_trf_state = CTRL_TRF_RX) then
        // It's the Data stage of a Control Write transfer
        USBCtrlTrfRxService()

//        // for control transfers only
//        if ctrl_trf_session_owner = MUID_HID then
//            // The only HID-class request with a host-to-device Data stage is Set_Report
//            HandleControlOutReport()
//        endif

        // Don't have to worry about overwriting _KEEP bit because if _KEEP was set,
        // TRNIF would not have been generated in the first place
      #if defined(INSTALL_MCHIP_USB_V13A_FIX)
        ep0Bo.ADR = addressof(CtrlTrfData)
        ep0Bo.Cnt = EP0_BUFF_SIZE
      #endif
        if (ep0Bo.Stat.DTS = 0) then
            ep0Bo.Stat._byte = _USIE or _DAT1 or _DTSEN
        else
            ep0Bo.Stat._byte = _USIE or _DAT0 or _DTSEN
        endif
    else    // the last OUT transaction must have been a status stage of a CTRL_TRF_TX
        //Prepare EP0 OUT for the next SETUP transaction
        USBPrepareForNextSetupTrf()
      #if defined(INSTALL_MCHIP_USB_V13A_FIX)
        ep0Bo.Cnt = EP0_BUFF_SIZE
        ep0Bo.ADR = addressof(SetupPkt)
        ep0Bo.Stat._byte = _USIE or _DAT0 or _DTSEN or _BSTALL
      #endif
    endif
end sub

//
//**************************************************************************
//Name    : USBCtrlTrfInHandler                                            *
//Purpose : This routine handles an IN transaction according to which      *
//        : control transfer state is currently active. A Set Address      *
//        : Request must not change the acutal address of the device until *
//        : the completion of the control transfer. The end of the control *
//        : transfer for Set Address. Request is an in transaction.        *
//        : Therefore it is necessary to service this unique situation when*
//        :  the condition is right. Macro mUSBCheckAdrPendingState is     *
//        : defined in usb9.h and its function is to specifically service  *
//        : this event.                                                    *
//**************************************************************************
//
sub USBCtrlTrfInHandler()                                                       'usbctrltr.c
    // Direction of data packet: device to host
    mUSBCheckAdrPendingState()         // Must check if in ADR_PENDING_STATE

    if (ctrl_trf_state = CTRL_TRF_TX) then
        // It's the Data stage of a Control Read transfer
        USBCtrlTrfTxService()

        //
        //********************************************************************
        //Bug Fix: May 14, 2007 (#F7 - Partial 2/4)
        //*********************************************************************
        //For a control transfer read, if the host tries to read more data
        //than what it has requested, the peripheral device should stall the
        //extra IN transactions and the status stage. Typically, a host does
        //not try to read more data than what it has requested. The original
        //firmware did not handle this situation. Instead of stalling extra
        //IN transactions, the device kept sending out zero length packets.
        //
        //This work around checks if a short IN packet has been sent or not.
        //If it has, the IN endpoint will be set to stall the next IN token.
        //If not, then the original endpoint setup code will be executed.
        //********************************************************************
        //
        if (short_pkt_status = SHORT_PKT_SENT) then
            // If a short packet has been sent, don't want to send any more,
            // stall next time if host is still trying to read
            ep0Bi.Stat._byte = _USIE or _BSTALL
            //*******************************************************************
        // **BUG FIX 1.3-3 add else block
        else
            if (ep0Bi.Stat.DTS = 0) then
                ep0Bi.Stat._byte = _USIE or _DAT1 or _DTSEN
            else
                ep0Bi.Stat._byte = _USIE or _DAT0 or _DTSEN
            endif
        endif
    else    // CTRL_TRF_RX
        // It's the Status stage of a Control Write transfer
        USBPrepareForNextSetupTrf()
    endif
end sub

//
//**************************************************************************
//Name    : USBCtrlEPService                                               *
//Require : USTAT is loaded with a valid endpoint address                  *
//Purpose : USBCtrlEPService checks for three transaction types that it    *
//        : knows how to service and services them.                        *
//        :    1. EP0 SETUP                                                *
//        :    2. EP0 OUT                                                  *
//        :    3. EP0 IN                                                   *
//        : It ignores all other types (i.e. EP1, EP2, etc.)               *
//**************************************************************************
//
// ********************************************************************
// Bug Fix: May 14, 2007
// ********************************************************************
function USBCtrlEPService() as byte
    dim setup_cnt as byte

    if (USTAT = EP00_OUT) then
        UIRBits.TRNIF = 0
        if (GetPID() = SETUP_TOKEN) then    // EP0 SETUP
            setup_cnt = 0
          #if defined(INSTALL_MCHIP_USB_V13A_FIX)
            //Check if the SETUP transaction data went into the CtrlTrfData buffer
            //If so, need to copy it to the SetupPkt buffer so that it can be
            //processed correctly by USBCtrlTrfSetupHandler()
            if (ep0Bo.ADR = addressof(CtrlTrfData)) then
                ep0Bo.ADR = addressof(SetupPkt)
                FSR0 = addressof(SetupPkt)
                FSR1 = addressof(CtrlTrfData)
                setup_cnt = sizeof(CTRL_TRF_SETUP)
                while (setup_cnt <> 0)
                    POSTINC0 = POSTINC1     // copy CtrlTrfData -> SetupPkt
                    dec(setup_cnt)
                end while
            endif
          #endif
            USBCtrlTrfSetupHandler()        //Handle the control transfer (parse the 8-byte SETUP command and figure out what to do)
        else                                // EP0 OUT
            USBCtrlTrfOutHandler()          //Handle the DATA transfer
        endif
    elseif (USTAT = EP00_IN) then           // EP0 IN
        UIRBits.TRNIF = 0
        USBCtrlTrfInHandler()
    else
        result = 0           // Return '0', if not an EP0 transaction
        exit
    endif
    result = 1
end function

//
//**************************************************************************
//Name    : USBModuleEnable                                                *
//Purpose : This routine enables the USB module. An end designer should    *
//        : never have to call this routine manually. This routine should  *
//        : only be called from USBCheckBusStatus().                       *
//**************************************************************************
//
sub USBModuleEnable()                                                           'usbdrv.c
    UCON = 0
    UIE = 0                                // Mask all USB interrupts
    UCONBits.USBEN = 1                     // Enable module & attach to bus
    usb_device_state = ATTACHED_STATE      // Defined in usbmmap.c & .h
end sub

//
//**************************************************************************
//Name    : USBModuleDisable                                               *
//Purpose : This routine disables the USB module. An end designer should   *
//        : never have to call this routine manually. This routine should  *
//        : only be called from USBCheckBusStatus().                       *
//**************************************************************************
//
// Note: When disabling the USB module, make sure the SUSPND bit (UCON<1>) is clear
// prior to clearing the USBEN bit. Clearing the USBEN bit when the module is in the
// suspended state may prevent the module from fully powering down
//
sub USBModuleDisable()                                                          'usbdrv.c
    UCONBits.SUSPND = 0                     // V1.4 - clear SUSPND first
    UCON = 0                                // Disable module & detach from bus
    UIE = 0                                 // Mask all USB interrupts
    usb_device_state = DETACHED_STATE       // Defined in usbmmap.c & .h
end sub

//
//**************************************************************************
//Name    : USBCheckBusStatus                                              *
//Purpose : This routine enables/disables the USB module by monitoring the *
//        : USB power signal.                                              *
//**************************************************************************
//
public sub USBCheckBusStatus()                                                  'usbdrv.c
    // Bus Attachment & Detachment Detection
    // usb_bus_sense is an i/o pin defined in io_cfg.h
    const
        USB_BUS_ATTACHED = 1,
        USB_BUS_DETACHED = 0

    if (usb_bus_sense = USB_BUS_ATTACHED) then      // Is USB bus attached?
        if (UCONBits.USBEN = 0) then                // Is the module off?
            USBModuleEnable()                       // Is off, enable it
        endif
    else
        if (UCONBits.USBEN = 1) then                // Is the module on?
            USBModuleDisable()                      // Is on, disable it
        endif
    endif

    // After enabling the USB module, it takes some time for the voltage
    // on the D+ or D- line to rise high enough to get out of the SE0 condition
    // The USB Reset interrupt should not be unmasked until the SE0 condition is
    // cleared. This helps preventing the firmware from misinterpreting this
    // unique event as a USB bus reset from the USB host
    if (usb_device_state = ATTACHED_STATE) then
        if (UCONBits.SE0 = 0) then
            UIR = 0                         // Clear all USB interrupts
            UIE = UIE_URSTIE or UIE_IDLEIE  // Unmask RESET & IDLE interrupts only
            usb_device_state = POWERED_STATE
        endif                               // else wait until SE0 is cleared
    endif
end sub

//
//**************************************************************************
//Name    : USBSoftDetach                                                  *
//Purpose : The device will have to be re-enumerated to function again     *
//        : USBSoftDetach electrically disconnects the device from the bus *
//        : This is done by stop supplying Vusb voltage to pull-up resistor*
//        : The pull-down resistors on the host side will pull both        *
//        : differential signal lines low and the host registers the event *
//        : as a disconnect.                                               *
//        : Since the USB cable is not physically disconnected, the power  *
//        : supply through the cable can still be sensed by the device. The*
//        : next time USBCheckBusStatus() function is called, it will      *
//        : reconnect the device back to the bus.                          *
//**************************************************************************
//
public sub USBSoftDetach()                                                             'usbdrv.c
    USBModuleDisable()
end sub

//
//**************************************************************************
//Name    : USBWakeFromSuspend                                             *
//Purpose :                                                                *
//**************************************************************************
//
sub USBWakeFromSuspend()                                                        'usbdrv.c
    // If using clock switching, this is the place to restore the original clock frequency
    UCONBits.SUSPND = 0
    UIEBits.ACTVIE = 0

    // V1.3 PIC18F2450/4450 Data Sheet Errata
    while (UIRBits.ACTVIF <> 0)
        UIRBits.ACTVIF = 0
    end while
end sub

//
//**************************************************************************
//Name    : USBProtocolResetHandler                                        *
//Require : A USB bus reset is received from the host                      *
//Purpose : Currently, this routine flushes any pending USB transactions.  *
//        : It empties OUT the USTAT FIFO. This action might not be        *
//        : desirable in some applications.                                *
//        : Once a USB bus reset is received from the host, this routine   *
//        : should be called. It resets the device address to zero,        *
//        : disables all non-EP0 endpoints, initializes EP0 to be ready for*
//        : default communication, clears all USB interrupt flags, unmasks *
//        : applicable USB interrupts, and reinitializes internal          *
//        : state-machine variables.                                       *
//**************************************************************************
//
inline sub Nop()
    asm
        Nop
    end asm
end sub

sub USBProtocolResetHandler()                                                   'usbdrv.bas
    UEIR = 0                                // Clear all USB error flags
    UIR = 0                                 // Clears all USB interrupts
    UEIE = UEIE_ENABLE_ALL                      // Unmask all USB error interrupts
    UIE = UIE_ENABLE_ALL and not(UIE_ACTVIE)    // Unmask all interrupts except ACTVIE

    UADDR = $00                             // Reset to default address
    mDisableAllEP()                         // Reset all non-EP0 UEPn registers
    UEP0 = EP_CTRL or HSHK_EN               // Init EP0 as a Ctrl EP, see usbdrv.h

    while (UIRBits.TRNIF = 1)               // Flush any pending transactions
        UIRBits.TRNIF = 0
        //
        //********************************************************************
        //Bug Fix: August 14, 2007
        //********************************************************************
        //Clearing the transfer complete flag bit, TRNIF, causes the SIE to
        //advance the FIFO. If the next data in the FIFO holding register is
        //valid, the SIE will reassert the interrupt within 6Tcy of clearing
        //TRNIF. If no additional data is preset, TRNIF will remain clear.
        //Additional nops were added in this fix to guarantee that TRNIF is
        //properly updated before being checked again.
        //********************************************************************
        //
        Nop() Nop() Nop()
        Nop() Nop() Nop()
    end while

    UCONBits.PKTDIS = 0            // Make sure packet processing is enabled
    USBPrepareForNextSetupTrf()    // Declared in usbctrltrf.c

  #if defined(INSTALL_MCHIP_USB_V13A_FIX)
    //Prepare EP0 OUT to receive the first SETUP packet
    ep0Bo.Cnt = EP0_BUFF_SIZE
    ep0Bo.ADR = addressof(SetupPkt)
    ep0Bo.Stat._byte = _USIE or _DAT0 or _DTSEN or _BSTALL
  #endif

    usb_stat.RemoteWakeup = 0      // Default status flag to disable
    usb_active_cfg = 0             // Clear active configuration
    usb_device_state = DEFAULT_STATE
end sub

//
//**************************************************************************
//Name    : USBRemoteWakeup                                                *
//Purpose : This function should be called by user when the device is waken*
//        : up by an external stimulus other than ACTIVIF. Please read the *
//        : note below to understand the limitations.                      *
//        : The modifiable section IN this routine should be changed to    *
//        : meet the application needs. Current implementation temporary   *
//        : blocks other functions from executing for a period of 1-13 ms  *
//        : depending on the core frequency. According to USB 2.0          *
//        : specification section 7.1.7.7,                                 *
//        :    "The remote wakeup device must hold the resume signaling    *
//        :     for at lest 1 ms but for no more than 15 ms."              *
//        : The idea here is to use a delay counter loop, using a common   *
//        : value that would work over a wide range of core frequencies.   *
//        :    ==========================================================  *
//        :    Core Freq(MHz)      MIP         RESUME Signal Period (ms)   *
//        :    ==========================================================  *
//        :        48              12          1.05                        *
//        :         4              1           12.6                        *
//        :    ==========================================================  *
//        : These timing could be incorrect when using code optimization   *
//        : or extended instruction mode, or when having other interrupts  *
//        : enabled. Make sure to verify using the MPLAB SIM's Stopwatch   *
//**************************************************************************
//
sub USBRemoteWakeup()                                                           'usbdrv.c
    // Check if RemoteWakeup function has been enabled by the host
    if (usb_stat.RemoteWakeup = 1) then
        //
        //********************************************************************
        //Bug Fix: August 14, 2007
        //********************************************************************
        //ACTVIF should be clear after exiting USBRemoteWakeup().
        //This was supposed to be done when calling USBWakeFromSuspend() from
        //USBRemoteWakeup(). However, this was not the case because when the
        //RESUME signal is sent, it causes the ACTVIF bit to set again.
        //By clearing the SUSPND bit only, it leaves the ACTVIE as one, and
        //when the RESUME signal is sent, it causes the ACTVIF to set.
        //Now both ACTVIE and ACTVIF are set, allowing the firmware to
        //properly see the event and call USBWakeFromSuspend() on the next
        //polling cycle of USBDriverService().
        //*******************************************************************
        //
        // Unsuspend USB module
        //USBWakeFromSuspend()             // Removed
        UCONBits.SUSPND = 0                // Added
        //*******************************************************************

        UCONBits.RESUME = 1                 // Start RESUME signaling

        // ****Modifiable Section****
        // Set RESUME line for 1-13 ms
        // ****End Modifiable Section****

        UCONBits.RESUME = 0
    endif
end sub

//
//**************************************************************************
//Name    : USBSuspend                                                     *
//Purpose :                                                                *
//**************************************************************************
//
sub USBSuspend()                                                                'usbdrv.c
    // NOTE: Do not clear UIRbits.ACTVIF here!
    // Reason:
    // ACTVIF is only generated once an IDLEIF has been generated.
    // This is a 1:1 ratio interrupt generation.
    // For every IDLEIF, there will be only one ACTVIF regardless of
    // the number of subsequent bus transitions.
    //
    // If the ACTIF is cleared here, a problem could occur when:
    // [       IDLE       ][bus activity ->
    // <--- 3 ms ----->     ^
    //                ^     ACTVIF=1
    //                IDLEIF=1
    //  #           #           #           #   (#=Program polling flags)
    //                          ^
    //                          This polling loop will see both IDLEIF=1 and ACTVIF=1
    //                          However, the program services IDLEIF first because ACTIVIE=0
    //                          If this routine clears the only ACTIVIF, then it can never get
    //                          out of the suspend mode
    UIEBits.ACTVIE = 1                     // Enable bus activity interrupt
    UIRBits.IDLEIF = 0
    UCONBits.SUSPND = 1                    // Put USB module in power conserve mode, SIE clock inactive

    // At this point the PIC can go into sleep,idle, or switch to a slower clock, etc
    // ****Modifiable Section****
//   PIR2Bits.USBIF = 0
//   INTCONBits.RBIF = 0
//   PIE2Bits.USBIE = 1                     // Set USB wakeup source
//   INTCONBits.RBIE = 1                    // Set sw2,3 wakeup source
//   asm                                   // Goto sleep
//       sleep
//   end asm
//   if INTCONBits.RBIF = 1 then            // Check if external stimulus
//      USBRemoteWakeup()                   // If yes, attempt RWU
//   endif
//   PIE2Bits.USBIE = 0
//   INTCONBits.RBIE = 0
    // ****End Modifiable Section****
end sub

//
//**************************************************************************
//Name    : USB_SOF_Handler                                                *
//Purpose : The USB host sends OUT a SOF packet to full-speed devices every*
//        : 1 ms. This interrupt may be useful for isochronous pipes. End  *
//        : designers should implement callback routine as necessary.      *
//**************************************************************************
//
sub USB_SOF_Handler()                                                           'usbdrv.c
    UIRBits.SOFIF = 0
    usb_Attached = (usb_device_state = CONFIGURED_STATE)
end sub

//
//**************************************************************************
//Name    : USBStallHandler                                                *
//Require : A STALL packet is sent to the host by the SIE.                 *
//Purpose : The STALLIF is set anytime the SIE sends OUT a STALL packet    *
//        : regardless of which endpoint causes it. A Setup transaction    *
//        : overrides the STALL function. A stalled endpoint stops stalling*
//        : once it receives a setup packet. In this case, the SIE will    *
//        : accept the Setup packet and set the TRNIF flag to notify the   *
//        : firmware. STALL function for that particular endpoint pipe will*
//        : be automatically disabled (direction specific).                *
//        : There are a few reasons for an endpoint to be stalled.         *
//        :    1. When a non-supported USB request is received.            *
//        :       Example: GET_DESCRIPTOR(DEVICE_QUALIFIER)                *
//        :    2. When an endpoint is currently halted.                    *
//        :    3. When the device CLASS specifies that an endpoint must    *
//        :       stall in response to a specific event.                   *
//        : UEPn.EPSTALL can be scanned to see which endpoint causes the   *
//        : stall event.                                                   *
//**************************************************************************
//
sub USBStallHandler()
    //
    //Does not really have to do anything here, even for the control endpoint
    //All BDs of Endpoint 0 are owned by SIE right now, but once a Setup Transaction
    //is received, the ownership for EP0_OUT will be returned to CPU.
    //When the Setup Transaction is serviced, the ownership for EP0_IN will then be
    //forced back to CPU by firmware
    //
    if (UEP0Bits.EPSTALL = 1) then
        //
        //********************************************************************
        //Bug Fix: August 14, 2007 (#F4)
        //********************************************************************
        //In a control transfer, when a request is not supported, all
        //subsequent transactions should be stalled until a new SETUP
        //transaction is received. The original firmware only stalls the
        //first subsequent transaction, then ACKs others. Typically, a
        //compliance USB host will stop sending subsequent transactions
        //once the first stall is received. In the original firmware,
        //function USBStallHandler() in usbdrv.c calls
        //USBPrepareForNextSetupTrf() when a STALL event occurred on EP0.
        //In turn, USBPrepareForNextSetupTrf() reconfigures EP0 IN and OUT
        //to prepare for the next SETUP transaction. The work around is not
        //to call USBPrepareForNextSetupTrf() in USBStallHandler().
        //
        //USBPrepareForNextSetupTrf();      // Removed

        //
        //********************************************************************
        //Bug Fix: August 14, 2007 (#F7 - Partial 4/4)
        //********************************************************************
        //For a control transfer read, if the host tries to read more data
        //than what it has requested, the peripheral device should stall the
        //extra IN transactions and the status stage. Typically, a host does
        //not try to read more data than what it has requested. The original
        //firmware did not handle this situation. Instead of stalling extra
        //IN transactions, the device kept sending out zero length packets.
        //
        //This work around checks to make sure that at least one extra IN
        //transaction is stalled before setting the OUT endpoint to stall the
        //status stage.
        //********************************************************************
        //
        if ((ep0Bo.Stat._byte = _USIE) and (ep0Bi.Stat._byte = (_USIE or _BSTALL))) then
            // Set ep0Bo to stall also
            ep0Bo.Stat._byte = _USIE or _DAT0 or _DTSEN or _BSTALL
        endif
        UEP0Bits.EPSTALL = 0              // Clear STALL status
    endif
    UIRBits.STALLIF = 0
end sub

//
//**************************************************************************
//Name    : USBErrorHandler                                                *
//Purpose : The purpose of this interrupt is mainly for debugging during   *
//        : development. Check UEIR to see which error causes the interrupt*
//**************************************************************************
//
sub USBErrorHandler()                                                           'usbdrv.c
    //
    //********************************************************************
    //Bug Fix: August 14, 2007
    //*********************************************************************
    //UERRIF bit is a read-only bit and cannot be cleared directly
    //The correct way is to clear UEIR register
    //********************************************************************
    //
    // UIRbits.UERRIF = 0           // Removed
    UEIR = 0
end sub

//
//**************************************************************************
//Name    : USBDriverService                                               *
//Purpose : This routine is the heart of this firmware. It manages all USB *
//        : interrupts. Device state transitions through the following     *
//        : stages: DETACHED -> ATTACHED -> POWERED -> DEFAULT ->          *
//        : ADDRESS_PENDING -> ADDRESSED -> CONFIGURED -> READY            *
//**************************************************************************
//
dim bTRNIFCount as byte

public sub USBDriverService()                                                   'usbdrv.c
    // Pointless to continue servicing if USB cable is not even attached
    if (usb_device_state = DETACHED_STATE) then
        exit
    endif

    // Task A: Service USB Activity Interrupt
    if ((UIRBits.ACTVIF = 1) and (UIEBits.ACTVIE = 1)) then
        USBWakeFromSuspend()
    endif

    // Pointless to continue servicing if the device is in suspend mode
    if (UCONBits.SUSPND = 1) then
        usb_Attached = false
        exit
    endif

    // Task B: Service USB Bus Reset Interrupt
    // When bus reset is received during suspend, ACTVIF will be set first,
    // once the UCONbits.SUSPND is clear, then the URSTIF bit will be asserted.
    // This is why URSTIF is checked after ACTVIF.
    // The USB reset flag is masked when the USB state is in DETACHED_STATE or
    // ATTACHED_STATE, and therefore cannot cause a USB reset event during these two states.

    if ((UIRBits.URSTIF = 1) and (UIEBits.URSTIE = 1)) then
        USBProtocolResetHandler()
    endif

    // Task C: Service other USB interrupts
    if ((UIRBits.IDLEIF = 1) and (UIEBits.IDLEIE = 1)) then
        USBSuspend()
    endif
    if ((UIRBits.SOFIF = 1) and (UIEBits.SOFIE = 1)) then
        USB_SOF_Handler()
    endif
    if ((UIRBits.STALLIF = 1) and (UIEBits.STALLIE = 1)) then
        USBStallHandler()
    endif
    if ((UIRBits.UERRIF = 1) and (UIEBits.UERRIE = 1)) then
        USBErrorHandler()
    endif

    // Pointless to continue servicing if the host has not sent a bus reset
    // Once bus reset is received, the device transitions into the DEFAULT
    // state and is ready for communication
    if (usb_device_state < DEFAULT_STATE) then
        exit
    endif

    // Task D: Servicing USB Transaction Complete Interrupt
    //********************************************************************
    //Bug Fix: August 14, 2007
    //********************************************************************
    bTRNIFCount = 0
    while (bTRNIFCount < 4)
        if ((UIRBits.TRNIF = 1) and (UIEBits.TRNIE = 1)) then
            // USBCtrlEPService only services transactions over EP0
            // It ignores all other EP transactions
            if (USBCtrlEPService() = 0) then      // If not an EP0 transaction, then clear TRNIF
                //
                //Other EP can be serviced later by responsible device class firmware.
                //Each device driver knows when an OUT or IN transaction is ready by
                //checking the buffer ownership bit.
                //An OUT EP should always be owned by SIE until the data is ready.
                //An IN EP should always be owned by CPU until the data is ready.
                //
                //Because of this logic, it is not necessary to save the USTAT value
                //of non-EP0 transactions
                //
                UIRBits.TRNIF = 0
                //
                //At least six Tcy are needed in between clearing UIR<TRNIF>
                //and when it becomes reasserted when the USTAT FIFO has more
                //than one entry in it
                //
                //No Nops are needed here because the number of instruction
                //cycles between clearing the TRNIF here to the next flag
                //check is longer than the required minimum of six Tcy due
                //to the for loop logic of bTRNIFCount. This is true even
                //when all optimization options in C18 are enabled.
                //
            endif
        else
            break
        endif
        inc(bTRNIFCount)
    end while

    // **BUG FIX 1.3-4: removed code after this point... it was a replication of
    // of the code prior to Task D bug fix above.

end sub

//
//**************************************************************************
// initialize the module
//**************************************************************************
//
// version 1.2
usb_Attached = false
ctrl_trf_state = WAIT_SETUP

// initialise descriptor pointers...
USB_CD_Ptr(0) = @cfg01
USB_CD_Ptr(1) = @cfg01          // configuration

USB_SD_Ptr(0) = @sd000          // language string
USB_SD_Ptr(1) = @sd001          // manufacturer string
USB_SD_Ptr(2) = @sd002          // product string
#if USB_HAVE_SERIAL_STRING
  USB_SD_Ptr(3) = @sd003        // serial number string
#endif

// HID defaults...
#if USB_USE_HID
  idle_rate = 0
  active_protocol = 0
#endif

// CDC defaults...
#if USB_USE_CDC
  cdc_tx_len = 0
  cdc_mem_type = _RAM
  cdc_trf_state = CDC_TX_READY
  CDCControlSignal._byte = 0
#endif // USB_USE_CDC
