//
//***************************************************************************
// Name    : usbmem.bas
// Author  : Jerry Messina
// Notice  : Copyright (c) 2011
//         : All Rights Reserved
// Date    : 4/7/2011
// Version : 1.4 initial release
// Notes   : USB Memory Map
//         : device-specific memory sizes and locations
// This file supports all current 18F USB devices from MPLAB 8.66/
// MPASM 5.40 (30 devices total) including:
// 18F2550, 18F4550, 18F2553, 18F4553, 18F2455, 18F4455, 18F2458, 18F4458,
// 18F2450, 18F4450, 18F14K50, 18F13K50, 18F67J50, 18F87J50, 18F66J50,
// 18F86J50, 18F66J55, 18F86J55, 18F65J50, 18F85J50, 18F24J50, 18F44J50,
// 18F25J50, 18F45J50, 18F26J50, 18F46J50, 18F26J53, 18F46J53, 18F27J53,
// 18F47J53
//***************************************************************************
//
module usbmem

#option SHOW_USB_BUILD_MSGS = false

//
//---------------------------------------------------------------------------
// USB device summary
//---------------------------------------------------------------------------
//
//  device      flash       ram             usb(BDT)        EPx endpoints****
//  18F2450     16K         512(000-1FF)    256(400-4FF)*** 64(0-63)    16
//  18F4450     16K         512(000-1FF)    256(400-4FF)    64(0-63)    16
//
//  18F2x50*    24-32K      2K(000-3FF)     1K(400-4FF)*    64(0-63)    16
//  18F4x50     24-32K      2K(000-3FF)     1K(400-4FF)*    64(0-63)    16
//
//  18F2xJ50    16-64K      3776(000-EBF)   all(400-4FF)**  64(0-63)    16
//  18F4xJ50    16-64K      3776(000-EBF)   all(400-4FF)**  64(0-63)    16
//
//  18F6xJ50    32-128K     3904(000-F3F)   all(400-4FF)**  64(0-63)    16
//  18F8xJ50    32-128K     3904(000-F3F)   all(400-4FF)**  64(0-63)    16
//
//  18F2xJ53    64-128K     3776(000-EBF)   all(D00-DFF)**  64(0-63)    16
//  18F4xJ53    64-128K     3776(000-EBF)   all(D00-DFF)**  64(0-63)    16
//
//  18F13K50    8K          256(000-0FF)    256(200-2FF)*** 32(0-31)    8
//  18F14K50    16K         512(000-1FF)    256(200-2FF)    32(0-31)    8
//
//
//  The ram sizes include the entire ram available (_maxram) excluding the SFR's.
//  For USB applications, this must be limited to account for the USB BDT and
//  buffer requirements.
//
//  * includes 2455, 2550, 4455, 4550, 2458, 2553, 4458, 4553
//    these parts differ only in flash size and 10/12-bit ADC
//  * BDT is at 400-4FF. USB data must be in the USB ram space (400-7FF).
//
//  ** for the 'J' series parts, only the BDT must reside in the USB page.
//     the USB data can be located anywhere in ram space, and can be any size up
//     to the SFR at the top of memory
//
//  *** for the 2450 and 13K/14K50 parts, the BDT and USB data must be contained
//      in the 256 byte USB bank (400/200). For the 1xK50, the BDT must fit into 
//      200h-27fh.
//
//  *** the EPx endpoints are bidirectional in/out pairs
//

//
//---------------------------------------------------------------------------
// SF and ram memory
//---------------------------------------------------------------------------
//
// The SF '_maxram' definitions found in the device *.bas files assume that all ram 
// can be used for general purposes, and are written for non-USB applications.
// This setting tells SF how much ram is available for it to allocate and use.
//
// The 18F USB pics have a special area of dual-port ram that is shared between the
// uC and the USB peripheral referred to as USB ram. This area appears in different
// locations and is different sizes in the various chips, but for each of the devices
// it is fixed in location (either at $200, $400, or $D00) and at least a portion of it 
// MUST be reserved for use specifically for things like the USB BDT and USB buffers. 
//
// To prevent SF from using the USB ram area, you must set '_maxram' to the address of 
// this fixed buffer. Unfortunately, on most chips this USB ram is located right in the 
// middle of the ram memory map, and so this limits the maximum amount of ram available 
// for use by SF since it can only use contiguous memory.
//
// In many cases, this leaves a large amount of the actual available ram unused. Even 
// large portions of the 'USB ram' are not normally required for USB usage, but this 
// too is located above the new '_maxram' setting, so it is also unavailable.
//
// To try and recoup some of this ram, USB Extended Ram was introducted. Extended Ram
// is ram (normally) located after the USB ram, and hence outside the reach of SF.
// It is managed by allocating blocks of memory manually using the 'absolute' keyword,
// at fixed locations in memory that might normally go otherwise unused.
//
// Since the original implementation, newer devices with different memory layouts
// have been introduced. Certain chips in the 'J' series can use ANY ram as USB buffer 
// ram, and others (like the 47J53) have moved the reserved USB block out of the middle 
// of the memory map and placed it at the top end of memory, which is much better.
//
// Some features of the code such as the HID ReadReport() and WriteReport() functions
// are written such that they require the use of Extended Ram, and are normally unavailable
// if Extended Ram is not being used.
//
// This implementation adds the ability to define and locate 'Extended Ram' anywhere
// in the memory map of the device as long as it is outside the block reserved for USB
// buffer ram, even BEFORE the USB ram. This allows you to use Extended Ram with devices 
// like the 47J53 (with it's USB ram at $D00) or the 14K50.
//

//
//---------------------------------------------------------------------------
// usage
//---------------------------------------------------------------------------
//
// #define USB_RAM_BASE = $400     // base address of the USB BDT ($200, $400, $D00)
// The BDT must reside here, along with the descriptors and usually the usb buffers
// Since SF can only use contiguous ram, this marks the end of ram available for use
// by SF, and _maxram may need to be adjusted (unless it's already less than this address)

// #define USB_RAM_MAX_SIZE = 1024                  // max available USB ram
// this is the max amount of ram available for USB usage (including the BDT page)
// some chips (such as the 'J' series) can use all ram, so there may actually
// more available, but this sets a reasonable limit

// #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT, in bytes
// this is the max size available for the BDT. Usually, this is limited to 256 bytes,
// but some parts (14K50) restrict it further. Normally, only a small portion of this
// is actually used by the BDT

// #define USB_RAM_SIZE = 1024     // USB ram, in bytes
// this is the actualy amount of ram reserved for use by the USB module itself.

// #variable USB_EXTENDED_RAM_BASE = $600        // if extended ram = true
// #variable USB_EXTENDED_RAM_SIZE = 1024        // USB extended ram, in bytes
// Extended ram is normally unused "usb" ram that can be reclaimed for general use,
// such as large buffers.
// You must have Extended Ram defined to use the HID ReadReport() and WriteReport()
// functions. It is typically located after the USB_RAM_BASE+USB_RAM_SIZE, and it must
// be manually allocated using the 'absolute' method since it is not (usually) in the
// contiguous ram area available for use by SF. However, it can be located anywhere
// outside of the memory block used by the USB module. If it is NOT located after
// the USB_RAM_BASE, then '_maxram' must be adjusted to account for it's presence.
//
// To use extended ram, set USB_EXTENDED_RAM_BASE to the desired memory address
// and set USB_EXTENDED_RAM_SIZE to the required size, such as
//    #define USB_EXTENDED_RAM_BASE = USB_RAM_BASE + USB_RAM_SIZE
//    #define USB_EXTENDED_RAM_SIZE = 512
//
// If not using Extended Ram, set USB_EXTENDED_RAM_SIZE to 0
//    #variable USB_EXTENDED_RAM_SIZE = 0         // disables extended ram usage
//
// You can also set '#option USB_DISABLE_EXTENDED_RAM = true' to disable extended ram
//
// In addition to Extended ram, there is normally extra ram available after the
// extended ram segment (USB_EXTENDED_RAM_BASE + USB_EXTENDED_RAM_SIZE). You can
// use this memory (up to the top of available ram) by directly allocating the
// memory yourself...
//
//  dim BigBuffer(1000) as byte absolute $0800
//  dim SecondBuffer(100) as byte absolute $0BE8
//
// The limitation of this is that you need to keep track of the location and size
// of each variable to calculate the correct address for the next one. So probably
// the best use for this RAM is for large arrays or structures.
//
//---------------------------------------------------------------------------
// USB memory definitions
//---------------------------------------------------------------------------
//
#if _device in (18F2450, 18F4450)
    // 768 bytes of total ram (includes 256 bytes of dual-port USB ram)
    // however, there is only 512 bytes available to SF ($000-$1FF)
    // since the USB ram is not contiguous
    #define USB_RAM_BASE = $400                     // BDT + descriptors + buffers
    #define USB_RAM_MAX_SIZE = 256                  // max available USB ram
    #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 256                    // USB ram, in bytes
    #variable USB_EXTENDED_RAM_SIZE = 128           // USB extended ram, in bytes (reduces SF ram)
    #variable USB_EXTENDED_RAM_BASE = $100          // extended ram addr (reduces SF ram)

#elseif _device in (18F2455, 18F2550, 18F4455, 18F4550, 18F2458, 18F2553, 18F4458, 18F4553)
    // 2K of ram + 1K of dual-port USB ram
    #define USB_RAM_BASE = $400                     // BDT + descriptors + buffers
    #define USB_RAM_MAX_SIZE = 1024                 // max available USB ram
    #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 512                    // USB ram, in bytes
    #variable USB_EXTENDED_RAM_SIZE = 512           // USB extended ram, in bytes
    #variable USB_EXTENDED_RAM_BASE = USB_RAM_BASE + USB_RAM_SIZE     // extended ram addr

#elseif _device in (18F24J50, 18F25J50, 18F26J50, 18F44J50, 18F45J50, 18F46J50)
    // 3776 bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $400                     // fixed BDT
    #define USB_RAM_MAX_SIZE = 3776                 // max available USB ram
    #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 512                    // USB ram, in bytes (can be to end of memory)
    #variable USB_EXTENDED_RAM_SIZE = 512           // USB extended ram, in bytes
    #variable USB_EXTENDED_RAM_BASE = USB_RAM_BASE + USB_RAM_SIZE     // extended ram addr

#elseif _device in (18F26J53, 18F27J53, 18F46J53, 18F47J53)
    // 3.8K bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $D00                     // fixed BDT
    #define USB_RAM_MAX_SIZE = 3776                 // max available USB ram
    #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 256+192                // USB ram, in bytes ($D00-$EBF)
    #variable USB_EXTENDED_RAM_SIZE = 512           // USB extended ram, in bytes
    #variable USB_EXTENDED_RAM_BASE = USB_RAM_BASE - 512       // $B00-CFF extended ram addr

#elseif _device in (18F65J50, 18F66J50, 18F66J55, 18F67J50, 18F85J50, 18F86J50, 18F86J55, 18F87J50)
    // 3904 bytes of total ram (USB ram can be anywhere)
    #define USB_RAM_BASE = $400                     // fixed BDT
    #define USB_RAM_MAX_SIZE = 3808                 // max available USB ram
    #define USB_BDT_MAX_SIZE = 256                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 512                    // USB ram, in bytes (can be to end of memory)
    #variable USB_EXTENDED_RAM_SIZE = 512           // USB extended ram, in bytes
    #variable USB_EXTENDED_RAM_BASE = USB_RAM_BASE + USB_RAM_SIZE       // extended ram addr

#elseif _device in (18F13K50, 18F14K50) then
    // 13K50 - 256 bytes of total ram (does not include 256 bytes of dual-port USB ram)
    // 14K50 - 768 bytes of total ram (includes 256 bytes of dual-port USB ram)
    // make sure _maxram from the device .bas file is correct for the part (some versions are not)
    #if _device in (18F13K50)
      #variable _maxram = $100              // 256 bytes of ram (can't use USB ram for SF)
    #elseif _device in (18F14K50)
      #variable _maxram = $300              // 768 bytes of ram
    #endif

    #define USB_RAM_BASE = $200                     // fixed BDT
    #define USB_RAM_MAX_SIZE = 256                  // max available USB ram
    #define USB_BDT_MAX_SIZE = 128                  // max size of the BDT

    // user-definable section
    #variable USB_RAM_SIZE = 256                    // USB ram, in bytes
    #if _device in (18F13K50)
      #variable USB_EXTENDED_RAM_SIZE = 0           // USB extended ram, in bytes
      #variable USB_EXTENDED_RAM_BASE = 0           // extended ram addr
    #elseif _device in (18F14K50)
      #variable USB_EXTENDED_RAM_SIZE = 128         // USB extended ram, in bytes (reduces SF ram)
      #variable USB_EXTENDED_RAM_BASE = $100        // extended ram addr (reduces SF ram)
    #endif

#else
  // add new devices here
  #error "unsupported device"
#endif


// reserve the USB BDT page (and up) from SF memory allocation
#if (USB_RAM_BASE < _maxram)
  #if (SHOW_USB_BUILD_MSGS)
    #warning "limiting _maxram available due to USB BDT"
  #endif
  #variable _maxram = USB_RAM_BASE
#endif

// check if user set '#option USB_DISABLE_EXTENDED_RAM = true' to disable extended ram
#if IsOption(USB_DISABLE_EXTENDED_RAM)
  #if (USB_DISABLE_EXTENDED_RAM = true)
    #if (SHOW_USB_BUILD_MSGS)
      #warning "disabling USB extended ram"
    #endif
    #variable USB_EXTENDED_RAM_SIZE = 0
  #elseif (USB_EXTENDED_RAM_SIZE = 0)       // user tried to enable extended ram, but no definition
    #error "you must set USB_EXTENDED_RAM_SIZE to use extended ram"
  #endif
#endif
#option USB_DISABLE_EXTENDED_RAM = false

// check to see if the user still has the old "USB_EXTENDED_RAM" option defined
// this is now unused, so issue a warning
#if IsOption(USB_EXTENDED_RAM)
  #warning "** obsolete USB_EXTENDED_RAM option detected. see usbmem.bas for new option **"
#endif

// if USB extended ram is used and it's located BEFORE the USB_RAM_BASE
// we may have to further restrict SF ram allocation
#if (USB_EXTENDED_RAM_SIZE > 0)
  #if (SHOW_USB_BUILD_MSGS)
    #warning "enabling USB extended ram"
  #endif
  #if (USB_EXTENDED_RAM_BASE < _maxram)
    #if (SHOW_USB_BUILD_MSGS)
      #warning "limiting _maxram available due to USB_EXTENDED_RAM_BASE"
    #endif
    #variable _maxram = USB_EXTENDED_RAM_BASE
  #endif
#endif

end module
